/**
 * 
 */
package sim.application;

import java.awt.BorderLayout;
import java.awt.EventQueue;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.IOException;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

import javax.swing.JButton;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JTextField;
import javax.swing.border.EmptyBorder;

import com.jogamp.opengl.GL;
import com.jogamp.opengl.GL2;
import com.jogamp.opengl.GLAutoDrawable;
import com.jogamp.opengl.GLCapabilities;
import com.jogamp.opengl.GLEventListener;
import com.jogamp.opengl.GLProfile;
import com.jogamp.opengl.awt.GLJPanel;
import com.jogamp.opengl.util.FPSAnimator;

import sim.application.util.SConfiguration;
import sim.exception.SConstructorException;
import sim.exception.SNoImplementationException;
import sim.graphics.SColor;
import sim.graphics.camera.SCamera;
import sim.graphics.camera.SVectorCamera;
import sim.listener.SListenerCamera;
import sim.math.SVector3d;
import sim.opengl.SGL;
import sim.opengl.SGLDisplay;
import sim.opengl.SGLTextureHandler;
import sim.opengl.SGLU;
import sim.opengl.SSimulationThread;
import sim.physics.SGravitationnalParticleSystem;
import sim.physics.SNumericalIntegrationMotion;
import sim.physics.SParticle;
import sim.physics.SParticleSystem;
import sim.physics.SPlanet;
import sim.physics.SStar;
import sim.readwrite.object.SMouseSetupRW;
import sim.readwrite.object.SPlanetRW;
import sim.readwrite.object.SSceneRW;
import sim.readwrite.object.SStarRW;
import sim.sound.SSoundUtil;
import sim.util.SBrowser;
import sim.util.SChronometer;
import sim.util.SLog;
import sim.util.SReadingException;
import javax.swing.SwingConstants;

/**
 * L'application <b>SIMOrbit</b> permet de visualiser des orbites de plantes en 3D.
 * 
 * @author Simon Vzina
 * @since 2023-09-27
 * @version 2023-10-25
 */
public class SIMOrbit extends JFrame implements GLEventListener {

  /**
   * 
   */
  private static final long serialVersionUID = -468249725965680475L;

  /**
   * La constante <b>WINDOWS_BUILDER_MODE</b> permet de configurer l'application afin de dterminer si l'application pourra activer ou non
   * les fonctionnalit de l'application intern windows builder.
   */
  private static final boolean WINDOWS_BUILDER_MODE = false;
   
  /**
   * La constante <b>DIGITS</b> reprsente le nombre de chiffres significatifs utiliss dans la reprsentation des nombres en notation scientifique tant gal  {@value}. 
   */
  private final static int DIGITS = 4;
  
  /**
   * La constante <b>SWITCH_TIME_FORMAT_VALUE</b> reprsente l'coulement requis  partir du quel nous allons changer le format de l'affichage (jour/heur/minute/seconde).
   */
  private final static double SWITCH_TIME_FORMAT_VALUE = 1000.0;
  
  /**
   * ...
   */
  private static final String CONFIGURATION_FILE_NAME = "orbit_configuration.cfg";
  
  /**
   * ...
   */
  private final SConfiguration configuration;
  
  /**
   * La variable <b>scene</b> reprsente la scne de cette simulation.
   */
  private SSceneRW scene;
  
  /**
   * La constante <b>TIME_STEP</b> reprsente le temps associ  la simulation.
   */
  private static final double DEFAULT_TIME_STEP = 0.01;
  
  /**
   * La camera qui pourra tre dplace si elle est connect au MouseEvent du JFrame.
   */
  private SListenerCamera camera;
  
  /**
   * ...
   */
  private List<SPlanet> celestial_body_list;
    
  /**
   * ...
   */
  private List<SStar> star_list;
  
  /**
   * ...
   */
  private List<LinkedList<SPlanet>> fading_list;
  
  /**
   * ...
   */
  private int fading_interval;
  
  /**
   * ...
   */
  private SParticleSystem sys = new SGravitationnalParticleSystem();
  
  
  /**
   * Le contrleur du rythme de l'affichage.
   */
  private final FPSAnimator fps;
  
  /**
   * Le contrleur de la simulation (s'il y en a une).
   */
  private final SSimulationThread simulator;
  
  /**
   * Le paramtre <b>contentPane</b> reprsente le panel principal de l'application.
   */
  private JPanel contentPane;
  
  /**
   * Champ d'affichage o il y aura affichage du temps coul depuis le dbut de l'excution de l'application.
   */
  private JTextField textFieldTime;
  
  /**
   * Champ d'affichage o il y aura affichage du temps coul de simulation.
   */
  private JTextField textFieldSimulationTime;
  private JTextField textFieldMomentum;
  private JTextField textFieldKineticEnergy;
  private JTextField textFieldAngularMomentum;
  private JTextField textFieldGavitationnalEnergy;
  private JTextField textFieldTotalEnergy;
  private JTextField textFieldScene;

  /**
   * Launch the application.
   */
  public static void main(String[] args)
  {
    // Instruction pour retirer des warning non pertinent.
    SGL.removeAccessWarning();
    
    // Lancement d'un thred pour excuteur le JFrame de l'application.
    EventQueue.invokeLater(new Runnable() {
      public void run() {
        try {
          SIMOrbit frame = new SIMOrbit();
          frame.setVisible(true);
        } catch (Exception e) {
          e.printStackTrace();
          System.exit(0);
        } 
      }
    });
  }

  /**
   * Create the frame.
   */
  public SIMOrbit() 
  {
    //-----------------------------------------
    // Construction du JFrame de l'application. 
    //-----------------------------------------
    super("Application pour tudier les orbites.");
    setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
    setBounds(100, 100, 900, 700);
    contentPane = new JPanel();
    contentPane.setBorder(new EmptyBorder(5, 5, 5, 5));
    setContentPane(contentPane);
    contentPane.setLayout(new BorderLayout(0, 0));
    
    //-------------------------------------------------------------------
    // Instruction pour la gestion des fonctionnalits systme de OpenGL.
    //-------------------------------------------------------------------
    GLProfile glProfile = GLProfile.getDefault();
    GLCapabilities glCapabilities = new GLCapabilities(glProfile);
    glCapabilities.setDoubleBuffered(true);
        
    GLJPanel glPanel = new GLJPanel(glCapabilities);    // construction du JPanel o sera affich le contenu OpenGL.
    glPanel.addGLEventListener(this);                   // Connecter le glPanel au vnement de type GL (repaint).
    
    //--------------------------------------------------------------------------
    // Structure de code pour permettre au plugin Design de fonctionner.
    // - En mode Design, il faut ajouter un JPanel ordinaire au centre du JFrame.
    // - En mode OpenGL, il faut ajouter un GLPanel au centre du JFrame.
    //--------------------------------------------------------------------------
    if(WINDOWS_BUILDER_MODE)
    {
      JPanel panelCenterToChangeForGLPanel = new JPanel();
      contentPane.add(panelCenterToChangeForGLPanel, BorderLayout.CENTER);
    }
    else
    {
      contentPane.add(glPanel, BorderLayout.CENTER);
    }
        
    //-----------------------------------------------------------------------------------------------
    // Construction de la camra par dfaut et connection des vnements de souris avec le glPanel et la camra.
    //-----------------------------------------------------------------------------------------------
    this.camera = new SListenerCamera(new SVectorCamera(new SVector3d(-5.0, 5.0, 0.0), SVector3d.ORIGIN, SVector3d.UNITARY_Z));
    glPanel.addMouseMotionListener(this.camera);
    glPanel.addMouseListener(this.camera);
    
    //-----------------------------------------------------------------------------------------
    // Construction des JPanel pour contenir le bouton et les TextField pour afficher du texte.
    //-----------------------------------------------------------------------------------------
    JPanel panelSouth = new JPanel();
    contentPane.add(panelSouth, BorderLayout.SOUTH);
    
    JButton btnPlay = new JButton("Action / Pause");
    btnPlay.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        actionPlayButton();
      }
    });
    panelSouth.setLayout(new GridLayout(3, 3, 0, 0));
    panelSouth.add(btnPlay);
    
    JPanel panelTime = new JPanel();
    panelSouth.add(panelTime);
    panelTime.setLayout(new GridLayout(2, 1, 0, 0));
    
    JLabel lblTime = new JLabel("Temps r\u00E9el :");
    lblTime.setHorizontalAlignment(SwingConstants.CENTER);
    panelTime.add(lblTime);
    
    textFieldTime = new JTextField();
    textFieldTime.setHorizontalAlignment(SwingConstants.CENTER);
    panelTime.add(textFieldTime);
    textFieldTime.setColumns(10);
    
    JPanel panelSimulationTime = new JPanel();
    panelSouth.add(panelSimulationTime);
    panelSimulationTime.setLayout(new GridLayout(2, 1, 0, 0));
    
    JLabel lblSimulationTime = new JLabel("Temps simul\u00E9 :");
    lblSimulationTime.setHorizontalAlignment(SwingConstants.CENTER);
    panelSimulationTime.add(lblSimulationTime);
    
    textFieldSimulationTime = new JTextField();
    textFieldSimulationTime.setHorizontalAlignment(SwingConstants.CENTER);
    panelSimulationTime.add(textFieldSimulationTime);
    textFieldSimulationTime.setColumns(10);
    
    JPanel panelMomentum = new JPanel();
    panelSouth.add(panelMomentum);
    panelMomentum.setLayout(new GridLayout(0, 1, 0, 0));
    
    JLabel lblNewLabelMomentum = new JLabel("Quantit\u00E9 de mouvement p (N/s) :");
    lblNewLabelMomentum.setHorizontalAlignment(SwingConstants.CENTER);
    panelMomentum.add(lblNewLabelMomentum);
    
    textFieldMomentum = new JTextField();
    textFieldMomentum.setHorizontalAlignment(SwingConstants.CENTER);
    panelMomentum.add(textFieldMomentum);
    textFieldMomentum.setColumns(10);
    
    JPanel panelNotUsed = new JPanel();
    panelSouth.add(panelNotUsed);
    
    JPanel panelAngularMomentum = new JPanel();
    panelSouth.add(panelAngularMomentum);
    panelAngularMomentum.setLayout(new GridLayout(0, 1, 0, 0));
    
    JLabel lblNewLabelAngularMomentum = new JLabel("Moment cin\u00E9tique (kg m^2/s^2) :");
    lblNewLabelAngularMomentum.setHorizontalAlignment(SwingConstants.CENTER);
    panelAngularMomentum.add(lblNewLabelAngularMomentum);
    
    textFieldAngularMomentum = new JTextField();
    textFieldAngularMomentum.setHorizontalAlignment(SwingConstants.CENTER);
    panelAngularMomentum.add(textFieldAngularMomentum);
    textFieldAngularMomentum.setColumns(10);
    
    JPanel panelKineticEnergy = new JPanel();
    panelSouth.add(panelKineticEnergy);
    
    JLabel lblNewLabelKineticEnergy = new JLabel("\u00C9nergie cin\u00E9tique (J) :");
    panelKineticEnergy.add(lblNewLabelKineticEnergy);
    
    textFieldKineticEnergy = new JTextField();
    textFieldKineticEnergy.setHorizontalAlignment(SwingConstants.CENTER);
    panelKineticEnergy.add(textFieldKineticEnergy);
    textFieldKineticEnergy.setColumns(10);
    
    JPanel panelGravitationnalEnergy = new JPanel();
    panelSouth.add(panelGravitationnalEnergy);
    
    JLabel lblNewLabelGravitationnalEnergy = new JLabel("\u00C9nergie gravitationnelle (J) :");
    panelGravitationnalEnergy.add(lblNewLabelGravitationnalEnergy);
    
    textFieldGavitationnalEnergy = new JTextField();
    textFieldGavitationnalEnergy.setHorizontalAlignment(SwingConstants.CENTER);
    panelGravitationnalEnergy.add(textFieldGavitationnalEnergy);
    textFieldGavitationnalEnergy.setColumns(10);
    
    JPanel panelTotalEnergy = new JPanel();
    panelSouth.add(panelTotalEnergy);
    
    JLabel lblNewLabelTotalEnergy = new JLabel("\u00C9nergie totale (J) :");
    panelTotalEnergy.add(lblNewLabelTotalEnergy);
    
    textFieldTotalEnergy = new JTextField();
    textFieldTotalEnergy.setHorizontalAlignment(SwingConstants.CENTER);
    panelTotalEnergy.add(textFieldTotalEnergy);
    textFieldTotalEnergy.setColumns(10);
    
    JPanel panel = new JPanel();
    contentPane.add(panel, BorderLayout.NORTH);
    
    JLabel lblNewLabelScene = new JLabel("Sc\u00E8ne :");
    panel.add(lblNewLabelScene);
    
    textFieldScene = new JTextField();
    textFieldScene.setHorizontalAlignment(SwingConstants.CENTER);
    panel.add(textFieldScene);
    textFieldScene.setColumns(20);
    
    JButton btnNewButtonLoading = new JButton("Chargement");
    btnNewButtonLoading.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        actionLoadingButton();
      }
    });
    panel.add(btnNewButtonLoading);
    
    JButton btnNewButtonBrowser = new JButton("Browser");
    btnNewButtonBrowser.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        actionBrowserButton();
      }
    });
    panel.add(btnNewButtonBrowser);
    
    //----------------------------------------------------------
    // Gestionnaire de l'affichage 
    //
    // C'est cet objet qui permet de relancer la mthode display
    // afin de faire le rafraichissement de l'cran
    //----------------------------------------------------------
    fps = new FPSAnimator(glPanel, 30, true); // 30 frames/sec
    fps.start();
    
    //----------------------------------------------------------
    // Gestionnaire de la simulation
    //
    // C'est cet objet qui va faire la mise  jour de la scne,
    // afin de raliser des simulations.
    //----------------------------------------------------------
    simulator = new SSimulationThread(this::simulation);
    simulator.setTimeStep(DEFAULT_TIME_STEP);
    simulator.start();
    simulator.setPlaying(false);
    
    // Initialisation du fichier de configuration et dtermination du nom du fichier de scne.
    try {
      
      configuration = new SConfiguration(CONFIGURATION_FILE_NAME);
      this.textFieldScene.setText(configuration.getReadDataFileName());
      
    }catch(IOException | SConstructorException e) {
      throw new RuntimeException("Erreur SIMOrbit 001 : Il y a eu une erreur lors du chargement du fichier de configuration.", e);
    }
    
    // Construction d'une scne null initialement.
    this.scene = null;
    
    // Allocation de la mmoire  la liste des plantes et des toiles.
    this.celestial_body_list = new ArrayList<SPlanet>();
    this.star_list = new ArrayList<SStar>();
    
    // Allocation de la mmoire pour les listes d'affichage en dgrad.
    this.fading_list = new ArrayList<LinkedList<SPlanet>>();
    
    // Excution de la premire initialisation
    loadScene();
  }

  /**
   * Mthode pour raliser les simulations dans la scene.
   * 
   * @param time_step L'coulement dans la simulation.
   */
  private void simulation(double time_step)
  {
    // Effectuer la simulation uniquement s'il est en action.
    if(this.simulator.isPlaying())
    {
      try {
    
        // Dplacer l'ensemble des particules du systme.
        this.sys.move(time_step, this.scene.getSimulationParameter().getNumericalAlgorithm());
        
      }catch(SNoImplementationException e) {
        
        // Il manque une implmentation, donc rien peut bouger en thorie.
        // Tenter de dplacer avec seulement Euler.
        moveEulerNoForce(time_step);
      }
    }
  }
  
  @Override
  public void display(GLAutoDrawable drawable)
  {   
    // Accs au fonctionnalit de OpenGL.
    GL2 gl = drawable.getGL().getGL2();
    
    // Effacer le buffer de couleur et de profondeur.
    gl.glClearColor(0,0,0,0);
    gl.glClear( GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT );
    
    //------------------------------------------------------------------------------------
    // Modification de la matrice du model view  partir des informations de la camra. //
    //------------------------------------------------------------------------------------
       
    // Informer OpenGL que c'est la matrice de model view qui sera modifie.
    gl.glMatrixMode(GL2.GL_MODELVIEW);
    
    // Effacer la matrice prcdente.
    gl.glLoadIdentity();
    
    // Modifier la matrice du model view avec une squence d'instruction openGL prprogramm dans GLU (GL_UTILITAIRE)
    SGLU.gluLookAt(gl, this.camera.getCamera());  
    
    // Effectuer le redimentionnement de la scne pour rduire la taille afin de permettre au Z-Buffer d'avoir plus de prcision. 
    if(this.scene != null)
      if(this.scene.getDisplayParameter().getRescalingFactor() != 1.0)
      {
        double factor = 1.0 / this.scene.getDisplayParameter().getRescalingFactor();
        gl.glScaled(factor, factor, factor);
      }
    
    //--------------------------------------------------------
    // Positionnement des sources de lumire tant les toiles
    //--------------------------------------------------------
      
    // Dfinir le dcalage du choix de la source de lumire  1 puisque la source #0 sera ddi pour l'clairage ambiante.
    int index_shift = 1;
            
    for(SStar s : this.star_list)
    {
      int index = GL2.GL_LIGHT0 + index_shift;
           
      // Activation de la source de lumire.
      gl.glEnable(index);
      
      // Dfinir la position et la couleur de la source de lumire.
      SGL.glLightfv(gl, index, s.getPosition(), SColor.BLACK, s.getColor(), s.getColor());
            
      // Identification de la prochaine source de lumire.
      index_shift++;
      
      // Limite du nombre de source de lumire utilisable sous openGL. 
      if(index_shift > SGL.MAX_LIGHT)
      {
        SLog.logWriteLine("Message SIMOrbit : Le nombre de lumire maximum  " + SGL.MAX_LIGHT + " a t atteint. Prsentement, il y a " + this.star_list.size() + " toiles dans la scne.");
        break;
      }
    }
    
    //--------------------------------------------------------------
    // Dessiner les plantes et toile (contenu dans la mme liste).
    //--------------------------------------------------------------
       
    //------------------------------------------------
    // tape 1 : Tout dessiner  la position actuelle.
    //------------------------------------------------
    for(SPlanet p : this.celestial_body_list)
      SGLDisplay.displaySphere(gl, SGLTextureHandler.loadTextureID(gl, p.getTextureName()), SGLDisplay.NO_TRANSPARENCY_VALUE, new SVector3d(p.getRay(), p.getRay(), p.getRay()), SVector3d.ZERO, p.getPosition());
  
    // Gestion de l'affichage avec dgrad.
    if(this.scene != null)
      if(this.scene.getDisplayParameter().getNbFading() > 0)
      {
        //------------------------------------------------
        // tape 2 : Afficher la liste de fading actuelle.
        //------------------------------------------------
        
        for(LinkedList<SPlanet> l : this.fading_list)
        {
          // Partir l'indice  1 afin que la 1er de la liste soit dj en attnuation de transparence.
          int index = 1;
          
          for(SPlanet p : l)
          {
            // Dfinir la transparence de la plante  afficher (rappel : 0 == 0% transparent)
            double transparency = index * 1.0 / this.scene.getDisplayParameter().getNbFading();
            
            SGLDisplay.displaySphere(gl, SGLTextureHandler.loadTextureID(gl, p.getTextureName()), transparency, new SVector3d(p.getRay(), p.getRay(), p.getRay()), SVector3d.ZERO, p.getPosition());
            index++;
          }
        }
      
        //---------------------------------------------------------------------------------------------------
        // tape 3 : Ajouter les plantes en position actuelle  la fading liste si admissible (si a bouge).
        //---------------------------------------------------------------------------------------------------
        if(this.simulator.isPlaying())
        {
          this.fading_interval++;
          
          // Valider qu'on est dans l'interval o l'on doit ajouter les images.
          if(this.fading_interval % this.scene.getDisplayParameter().getFadingInterval() == 0)
          {
            // Ajouter la plante active  la liste des dgrad.
            for(int i = 0; i < this.celestial_body_list.size(); i++)
              this.fading_list.get(i).addFirst(this.celestial_body_list.get(i).copy());
            
            // Vrifier si les listes ont dpasses la taille maximale.
            for(LinkedList<SPlanet> l : this.fading_list)
              if(l.size() > this.scene.getDisplayParameter().getNbFading())
                l.removeLast();
            
            // Repartir le compteur  zro.
            this.fading_interval = 0;
          }
        }
        
      }
    
    // Permuttation des buffer images (rendre l'affichage plus harmonieux)
    drawable.swapBuffers();
    
    //--------------------------------------------
    // Mise  jour des champ de texte (TextField).
    //--------------------------------------------
    updateTextField();
  }

  @Override
  public void dispose(GLAutoDrawable arg0)
  {
    // il n'y a rien  faire ...
  }

  @Override
  public void init(GLAutoDrawable drawable) 
  {
    // Accs au fonctionnalit de OpenGL.
    GL2 gl = drawable.getGL().getGL2();
    
    // Effacer les couleurs initiale.
    gl.glClearColor(0, 0, 0, 0);
    
    //-----------------------------------------------------------
    // ACTIVATION DE DIFFRENTE ALGORITHME DE CALCUL GRAPHIQUE //
    //-----------------------------------------------------------
    
    // Activer le test de profondeur pour afficher les triangles prs de la camra (pas dans l'ordre d'affichage).
    gl.glEnable(GL.GL_DEPTH_TEST);
        
    // Activer le test de la normale face  la camra (culling). Autrement, le polygone ne sera pas affich.
    gl.glCullFace(GL.GL_FRONT_FACE);
    
    // Activer le mode d'interpolation des couleurs sur les surface des triangles.
    gl.glShadeModel(GL2.GL_SMOOTH);  // version avec normale  la surface interpole.
           
    // Activer le mode clairage pour l'algorithme d'illumination.
    gl.glEnable(GL2.GL_LIGHTING);
    
    // Activer le mode couleur des matriaux pour l'algorithme d'illumination.
    gl.glEnable(GL2.GL_COLOR_MATERIAL);
    
    // Activer le mode "normalize" (que je ne sais pas pourquoi pour l'instant ...)
    gl.glEnable(GL2.GL_NORMALIZE);
    
    //---------------------------------------------------------------------------
    // ACTIVATION ET DFINITION DES PROPRITS DE LA SOURCE #0 : LUMIRE AMBIANTE
    //---------------------------------------------------------------------------
   
    // Activer la source de lumire #0.
    gl.glEnable(GL2.GL_LIGHT0);
    
    // Dfinition des paramtres de la source de lumire.
    SGL.glLightfv(gl, GL2.GL_LIGHT0, SVector3d.ORIGIN, SColor.WHITE.multiply(0.1), SColor.BLACK, SColor.BLACK);
  }

  @Override
  public void reshape(GLAutoDrawable drawable, int x, int y, int width, int height )
  {
    // Avoir accs au fonctionnalit de OpenGL.
    final GL2 gl = drawable.getGL().getGL2();
    
    // Dfinir la taille de la fente de vue.
    final float aspectRatio = ( float ) width / ( float ) height;
    gl.glViewport( 0, 0, width, height );
    
    // Mentionner  OpenGL que c'est la matrice de projection qui sera modifie.
    gl.glMatrixMode( GL2.GL_PROJECTION );
    
    // Charger la matrice identit (et effacer la matrice prcdente).
    gl.glLoadIdentity();
    
    // Utiliser les fonctionnalits de GLU pour modifier la matrice de projection avec les effets de perspective.
    SCamera c = this.camera.getCamera();
    SGLU.gluPerspective(gl,c.getViewAngle(), aspectRatio, c.getZNear(), c.getZFar());
  }
    
  /**
   * Mthode permettant de modifier le mode de la simulation (en action ou en pause).
   */
  private void actionPlayButton()
  {
    // Production d'un son pour dsigner si le systme de met en action ou en pause.
    try {
      
      // Chargement pour la fermeture de la simulation.
      if(this.simulator.isPlaying())
        SSoundUtil.playWaveFile("mixkit-arcade-retro-game-over-213.wav", false);
      else
        SSoundUtil.playWaveFile("mixkit-retro-game-notification-212.wav", false);
      
    }catch(IllegalArgumentException e) {
      // On ne gnre pas le son, mais on fait rien de plus ...
    }
    
    // Permutter l'tat de la simulation.
    this.simulator.switchPlaying();
  }
  
  /**
   * Mthode permettant de metter  jour les contenus des TextField.
   */
  private void updateTextField()
  {
    // Champ du temps rel requis pour la simulation.
    this.textFieldTime.setText(String.format("%.3f", this.simulator.getRealTime()) + " s");
    
    // Champ du temps de simulation.
    if(this.simulator.getSimulationTime() < SWITCH_TIME_FORMAT_VALUE)
      this.textFieldSimulationTime.setText(String.format("%.3f", this.simulator.getSimulationTime()) + " s");
    else
      this.textFieldSimulationTime.setText(SChronometer.secToTimeFormatYDHHMMSS((int)this.simulator.getSimulationTime()));
      
    // Champ du momentum.
    try {
      this.textFieldMomentum.setText(this.sys.momentum().toString(DIGITS));
    }catch(SNoImplementationException e) {
      this.textFieldMomentum.setText("Non implment.");
    }
    
    // Champ du angular momentum.
    try {
      this.textFieldAngularMomentum.setText(this.sys.angularMomentum().toString(DIGITS));
    }catch(SNoImplementationException e) {
      this.textFieldAngularMomentum.setText("Non implment.");
    }
    
    // Champ du kinetic energy.
    try {
      this.textFieldKineticEnergy.setText(String.format( "%." + DIGITS + "e", this.sys.kineticEnergy()));
    }catch(SNoImplementationException e) {
      this.textFieldKineticEnergy.setText("Non implment.");
    }
   
    // Champ du gravitationnal energy.
    try {
      this.textFieldGavitationnalEnergy.setText(String.format( "%." + DIGITS + "e", this.sys.potentialEnergy()));
    }catch(SNoImplementationException e) {
      this.textFieldGavitationnalEnergy.setText("Non implment.");
    }
    
    // Champ du total energy.
    try {
      this.textFieldTotalEnergy.setText(String.format( "%." + DIGITS + "e", this.sys.totalEnergy()));
    }catch(SNoImplementationException e) {
      this.textFieldTotalEnergy.setText("Non implment.");
    }
  }
    
  /**
   * Mthode pour faire la lecture de la scne.
   */
  private void loadScene()
  {
    try {
      
      // Lecture de la nouvelle scne temporairement.
      SSceneRW tmp_scene = new SSceneRW(this.textFieldScene.getText());
      tmp_scene.read();
      
      // Si aucune exception n'est lance, c'est que la scne sera adquate.
      this.simulator.reset();
      deleteScene();
      this.scene = tmp_scene;
          
      // Construire la camra.
      if(this.scene.getCameraList().isEmpty())
        SLog.logWriteLine("Message SIMOrbit : Attention, il n'y a pas de camra de dfini dans la scne.");
      else
        this.camera.setCamera(this.scene.getCameraList().get(0).toSVectorCamera());
      
      // Configuration du mouvement de la sourie avec la camra.
      SMouseSetupRW mouse_setup = this.scene.getMouseSetup();
      this.camera.setMovementSensibility(mouse_setup.getMovementSensibility() * mouse_setup.getDeltaMovement());
      this.camera.setRotationSensibility(mouse_setup.getRotationSensibility() * mouse_setup.getDeltaRotation());
          
      // Construire les plantes.
      for(SPlanetRW p : this.scene.getPlanetList())
        this.celestial_body_list.add(p.toSPlanet());
      
      // Construire les toiles. Mettre dans la liste des toiles et des plantes.
      for(SStarRW s : this.scene.getStarList())
      {
        SStar star = s.toSStar();
        
        this.star_list.add(star);
        this.celestial_body_list.add(star);
      }
      
      // Remplir le systme de particle (avec plante et toile)
      for(SPlanet p : celestial_body_list)
        this.sys.add(p.getParticle());
         
      // Construire les listes pour affichage en dgrad.
      for(int i = 0; i < celestial_body_list.size(); i++)
        this.fading_list.add(new LinkedList<SPlanet>());
      
      // Dfinir des paramtres du simulateur.
      this.simulator.setTimeStep(this.scene.getSimulationParameter().getTimeStep());
      this.simulator.setNbStep(this.scene.getSimulationParameter().getNbStep());
      this.fading_interval = 0;
      
      // criture de la scne (pour valider l'interprtation du fichier.
      try {
        this.scene.write(configuration.getWriteDataFileName());
      }catch(IOException e) {
        throw new RuntimeException("Erreur SIMOrbit 002 : Erreur d'criture de la scne.", e);
      }
      
      // Production d'un son signifiant que a fonctionn.
      SSoundUtil.playWaveFile("load_scene.wav", false);
      
      // criture dans la console des informations de la scne.
      SLog.logWriteLine("Message SIMOrbit : La scne en lecture est '"+ this.scene.getFileName() + "'.");
      
      for(SPlanet p : this.celestial_body_list)
        SLog.logWriteLine("Astre : " + p.toString());
      
      SLog.logWriteLine();
      
    }catch(SReadingException e) {
      SLog.logWriteLine("Message SIMOrbit : La scne '" + this.textFieldScene.getText() + "' ne peut pas tre charge, car il y a eu une erreur de lecture.");
      SLog.logException(e);
      SLog.logWriteLine("Message SIMOrbit : La scne actuelle sera conserve.");
      
      // Production d'un son signifiant que a n'a pas fonctionn.
      SSoundUtil.playWaveFile("chord.wav", false);
    }
    
  }
  
  /**
   * Mthode pour retirer les informations de la la scne.
   */
  private void deleteScene()
  {
    // quivalent  un destructeur.
    this.scene = null;
    
    this.sys.clear();
    this.star_list.clear();
    this.celestial_body_list.clear();
    this.fading_list.clear();
  }
  
  /**
   * Mthode pour dplacer l'ensemble des particules avec l'algorithme de Euler sans force.
   * 
   * @param dt Le temps coul dans le dplacement.
   */
  private void moveEulerNoForce(double dt)
  {
    try {
     
      // Dplacer l'ensemble des particules du systme.
      for(int i = 0; i < sys.getNbParticles(); i++)
      {
        SParticle p = sys.getParticle(i);
        p.setKineticComponent(SNumericalIntegrationMotion.euler(p.getKineticComponent(), SVector3d.ZERO, dt));
      }
    }catch(SNoImplementationException e) {
      // Sans implmentation de Euler, rien va bouger !
    }
  }
  
  /**
   * Mthode pour l'action du bouton de chargement de scne.
   */
  private void actionLoadingButton()
  {
    loadScene();
  }
  
  /**
   * Mthode pour l'action du bouton de la recherche du nom de la scne.
   */
  private void actionBrowserButton()
  {
    this.textFieldScene.setText(SBrowser.fileChooser());
  }
      
}