/**
 * 
 */
package sim.math;

import java.util.Arrays;

import sim.util.SArrays;

/**
 * La classe <b>SMatrix3x3</b> reprsentant une matrice de '3' lignes et '3' colonnes. 
 * Les composantes de la matrice respectent la notation suivante :
 * <p>M[i][j] = </p> 
 * <p>| m00  m01  m02 |</p>
 * <p>| m10  m11  m12 |</p>
 * <p>| m20  m21  m22 |</p>
 *  
 * Puisque l'implmentation de cette matrice se fera  l'aide d'un tableau  une dimension, on retrouve l'indexation suivante :
 * <p>M[i][j] = </p> 
 * <p>|  0    1    2   |</p>
 * <p>|  3    4    5   |</p>
 * <p>|  6    7    8   |</p>
 * 
 * @author Simon Vezina
 * @since 2020-05-05
 * @version 2021-12-27
 */
public class SMatrix3x3 {

  /**
   * La constance <b>EMPTY</b> reprsente le tableau d'une matrice vide.
   */
  private static final double[] EMPTY_ARRAY = { 
                      0.0, 0.0, 0.0, 
                      0.0, 0.0, 0.0,
                      0.0, 0.0, 0.0
                      };                                       
  
  /**
   * ...
   */
  private static final double[] POSITIVE_INFINITY_ARRAY = {
      Double.POSITIVE_INFINITY, Double.POSITIVE_INFINITY, Double.POSITIVE_INFINITY,
      Double.POSITIVE_INFINITY, Double.POSITIVE_INFINITY, Double.POSITIVE_INFINITY,
      Double.POSITIVE_INFINITY, Double.POSITIVE_INFINITY, Double.POSITIVE_INFINITY,
      };
  
  /**
   * ...
   */
  private static final double[] NEGATIVE_INFINITY_ARRAY = {
      Double.NEGATIVE_INFINITY, Double.NEGATIVE_INFINITY, Double.NEGATIVE_INFINITY,
      Double.NEGATIVE_INFINITY, Double.NEGATIVE_INFINITY, Double.NEGATIVE_INFINITY,
      Double.NEGATIVE_INFINITY, Double.NEGATIVE_INFINITY, Double.NEGATIVE_INFINITY,
      };
  
  /**
   * La constance <b>IDENTITY</b> reprsente le tableau d'une matrice identit.
   */                                       
  private static final double[] IDENTITY_ARRAY = { 
                    1.0, 0.0, 0.0,
                    0.0, 1.0, 0.0,
                    0.0, 0.0, 1.0 
                    };
  
  /**
   * La constante <b>ZERO</b> reprsente la matrice zro (contenant uniquement des zros).
   */
  public static final SMatrix3x3 ZERO = new SMatrix3x3(EMPTY_ARRAY);
  
  /**
   * La constante <b>POSITIVE_INFINITY</b> reprsente la matrice ayant des termes infinis positifs pour l'ensemble de ses composants.
   */
  public static final SMatrix3x3 POSITIVE_INFINITY = new SMatrix3x3(POSITIVE_INFINITY_ARRAY);
  
  /**
   * La constante <b>NEGATIVE_INFINITY</b> reprsente la matrice ayant des termes infinis ngatifs pour l'ensemble de ses composants.
   */
  public static final SMatrix3x3 NEGATIVE_INFINITY = new SMatrix3x3(NEGATIVE_INFINITY_ARRAY);
  
  /**
   * La constante <b>IDENTITY</b> reprsente la matrice identit.
   */
  public static final SMatrix3x3 IDENTITY = new SMatrix3x3(IDENTITY_ARRAY);
  
  //-------------
  // VARIABLES //
  //-------------
  
  /**
   * La variable 'matrix' correspond  la matrice 4x4 contenant les 16 lments.
   */
  private final double[] matrix;
  
  //-----------------
  // CONSTRUCTEURS //
  //-----------------
  
  /**
   * Constructeur de la matrice nulle (tous les lments M[i][j] seront nuls.)  
   */
  public SMatrix3x3()
  {
    this(EMPTY_ARRAY);
  }
  
  /**
   * ...
   * 
   * @param m00
   * @param m01
   * @param m02
   * @param m10
   * @param m11
   * @param m12
   * @param m20
   * @param m21
   * @param m22
   */
  public SMatrix3x3(double m00, double m01, double m02,
                    double m10, double m11, double m12,
                    double m20, double m21, double m22)
  {
    this.matrix = new double[]{ 
            m00,    m01,    m02,
            m10,    m11,    m12,
            m20,    m21,    m22
            };
  }
  
  /**
   * Constructeur d'une matrice avec tableau des 16 composantes de la matrice.
   * 
   * @param m Le tableau des 16 composantes de la matrice.
   */
  private SMatrix3x3(double[] m)
  {
    this.matrix = m;
  }
  
  /**
   * Mthode pour obtenir une composante de la matrice en indice i de la matrice o i correspond  la position d'un lment organis dans un tableau de valeur entre 0 et 8.
   *
   * @return La composante en position i dans le tableau des valeurs de la matrice.
   * @throws ArrayIndexOutOfBoundsException Si l'accs dans le tableau de la matrice n'est pas adquat.
   */
  public double get(int i) throws ArrayIndexOutOfBoundsException
  {
    return this.matrix[i];
  }
  
  /**
   * Mthode pour obtenir une valeur m[i,j] de la matrice. 
   * 
   * @param i L'indice de la ligne de la matrice.
   * @param j L'indice de la colonne de la matrice.
   * @return L'lment m[i,j] de la matrice.
   * @throws ArrayIndexOutOfBoundsException Si la coordonne (i,j) n'est pas  l'intrieure de (0,0)  (2,2) pour une matrice 3x3.
   */
  public double get(int i, int j) throws ArrayIndexOutOfBoundsException
  {
    if(i < 0 || j < 0)
      throw new ArrayIndexOutOfBoundsException("Erreur SMatrix3x3 001 : L'accs (i,j) = (" + i + "," + j + ") contient une valeur ngative.");
        
    return this.get(i*3 + j);
  }
  
  /**
   * Mthode qui effecteur l'addition de deux matrices.
   * 
   * @param M La matrice  additionner.
   * @return La matrice correspondant  la somme des deux matrices.
   */
  public SMatrix3x3 add(SMatrix3x3 M)
  {
    // Cas particulier considr :
    if(this == SMatrix3x3.ZERO && M == SMatrix3x3.ZERO)
      return SMatrix3x3.ZERO;
          
    return new SMatrix3x3(SArrays.add(this.matrix, M.matrix));
  }
  
  /**
   * Mthode qui effecteur la soustracteur de deux matrices.
   * 
   * @param M La matrice  soustraire.
   * @return La matrice correspondant  la soustraction des deux matrices.
   */
  public SMatrix3x3 substract(SMatrix3x3 M)
  {
    // Cas particulier considr :
    if(this == SMatrix3x3.ZERO && M == SMatrix3x3.ZERO)
      return SMatrix3x3.ZERO;
    
    return new SMatrix3x3(SArrays.substract(this.matrix, M.matrix));
  }
  
  /**
   * Mthode qui effectue la multiplication de la matrice par un <u>scalaire</u>.
   * 
   * @param a La valeur  multiplier.
   * @return La matrice multiplie par par un scalaire.
   */
  public SMatrix3x3 multiply(double a)
  {
    // Cas particulier considr :
    if(this == SMatrix3x3.ZERO || this == SMatrix3x3.POSITIVE_INFINITY || this == SMatrix3x3.NEGATIVE_INFINITY)
      return this;
    
    return new SMatrix3x3(SArrays.multiply(this.matrix, a));
  }
  
  /**
   * Mthode qui effectuer la <b>multiplication</b> de la matrice avec un <u>vecteur</u>.
   * 
   * @param v Le vecteur  multiplier.
   * @return Le vecteur rsultat de la multiplication de la matrice avec le vecteur.
   */
  public SVector3d multiply(SVector3d v)
  {
    return new SVector3d( 
        
        matrix[0]*v.getX() + matrix[1]*v.getY() + matrix[2]*v.getZ(),
        matrix[3]*v.getX() + matrix[4]*v.getY() + matrix[5]*v.getZ(),
        matrix[6]*v.getX() + matrix[7]*v.getY() + matrix[8]*v.getZ()   
        
        );
  }
  
  /**
   * Mthode qui effectue le produit de deux matrices.
   * 
   * @param M La matrice  multiplier.
   * @return Le produit des deux matrices.
   */
  public SMatrix3x3 multiply(SMatrix3x3 M)
  {
    // Cas particulier considr :
    if(this == SMatrix3x3.ZERO && M == SMatrix3x3.ZERO)
      return SMatrix3x3.ZERO;
    
    
    return new SMatrix3x3(  
        
        // Premiere ligne
        this.matrix[0]*M.matrix[0] + this.matrix[1]*M.matrix[3] + this.matrix[2]*M.matrix[6],         
        this.matrix[0]*M.matrix[1] + this.matrix[1]*M.matrix[4] + this.matrix[2]*M.matrix[7],         
        this.matrix[0]*M.matrix[2] + this.matrix[1]*M.matrix[5] + this.matrix[2]*M.matrix[8],         
        
        // Deuxieme ligne
        this.matrix[3]*M.matrix[0] + this.matrix[4]*M.matrix[3] + this.matrix[5]*M.matrix[6],         
        this.matrix[3]*M.matrix[1] + this.matrix[4]*M.matrix[4] + this.matrix[5]*M.matrix[7],         
        this.matrix[3]*M.matrix[2] + this.matrix[4]*M.matrix[5] + this.matrix[5]*M.matrix[8],         
        

        // Troisieme ligne
        this.matrix[6]*M.matrix[0] + this.matrix[7]*M.matrix[3] + this.matrix[8]*M.matrix[6],         
        this.matrix[6]*M.matrix[1] + this.matrix[7]*M.matrix[4] + this.matrix[8]*M.matrix[7],         
        this.matrix[6]*M.matrix[2] + this.matrix[7]*M.matrix[5] + this.matrix[8]*M.matrix[8]         
           
        );
  }
  
 
  /**
   * <p>Mthode pour obtenir la matrice identit I.</p> 
   * <p>Soit une matrice identit I et une
   * matrice A quelconque, alors</p> 
   * 
   *  <ul> A = A*I = I*A </ul>
   *  <p>ce qui rend la <b>matrice identit commutative</b>.</p>
   * 
   * @return La matrice identit I.
   */
  public static SMatrix3x3 identity()
  {
    return IDENTITY;
  }
  
  /**
   * ...
   * 
   * @return
   */
  public static SMatrix3x3 zero()
  {
    return ZERO;
  }
  
  /**
   * ...
   * 
   * @return
   */
  public static SMatrix3x3 positiveInfinity()
  {
    return POSITIVE_INFINITY;
  }
  
  /**
   * ...
   * 
   * @return
   */
  public static SMatrix3x3 negativeInfinity()
  {
    return NEGATIVE_INFINITY;
  }
  
  /**
   * Mthode pour obtenir la matrice de produit vectoriel associ  un vecteur.
   * Il est important de prciser que cette proprit est valide uniquement si l'on multiplie cette matrice avec un vecteur.
   * 
   * @param A Le vecteur des paramtres  mettre en matrice.
   * @return La matrice de produit vectoriel.
   */
  public static SMatrix3x3 skewSymmetric(SVector3d A)
  {
    return new SMatrix3x3(0.0, -A.getZ(), A.getY(),
                          A.getZ(), 0.0, -A.getX(),
                          -A.getY(), A.getX(), 0.0);
  }
  
  /**
   * Mthode pour obtenir la matrice transpose de la matrice qui lance l'appel de cette mthode.
   * La matrice transpose B s'obtient  partir d'une matrice A grce au calcul
   *
   * <ul> B[j][i] = A[i][j]</ul>
   * 
   * @return La matrice transpose.
   */
  public SMatrix3x3 transpose()
  {
    // Cas particulier considr :
    if(this == SMatrix3x3.ZERO || this == SMatrix3x3.POSITIVE_INFINITY || this == SMatrix3x3.NEGATIVE_INFINITY)
      return this;
    
    return new SMatrix3x3(
        
        this.matrix[0],  this.matrix[3],  this.matrix[6], 
        this.matrix[1],  this.matrix[4],  this.matrix[7], 
        this.matrix[2],  this.matrix[5],  this.matrix[8]
      
        );
  }
  
  /**
   * Mthode pour obtenir le dterminant de la matrice.
   * 
   * Rfrence : https://ncalculators.com/matrix/inverse-matrix.htm
   * 
   * @return Le dterminant de la matrice.
   */
  public double determinant()
  {
    return 
          this.matrix[0]*this.matrix[4]*this.matrix[8]
        + this.matrix[1]*this.matrix[5]*this.matrix[6]
        + this.matrix[2]*this.matrix[3]*this.matrix[7]
        - this.matrix[2]*this.matrix[4]*this.matrix[6]
        - this.matrix[0]*this.matrix[5]*this.matrix[7]
        - this.matrix[1]*this.matrix[3]*this.matrix[8];
  }
  
  /**
   * Mthode pour obtenir l'inverse de la matrice.
   * 
   * @return La matrice inverse.
   * @throws IllegalArgumentException Si la matrice ne peut pas tre inverse (dterminant nul).
   */
  public SMatrix3x3 inverse() throws IllegalArgumentException
  {
    // Cas particulier : L'inverse de la matrice zro sera la matrice infinie.
    if(this == SMatrix3x3.ZERO)
      return SMatrix3x3.POSITIVE_INFINITY;
    
    // Cas particuler : L'inverse de la matrice infinie sera la matrice zro.
    if(this == SMatrix3x3.POSITIVE_INFINITY || this == SMatrix3x3.NEGATIVE_INFINITY)
      return SMatrix3x3.ZERO;
        
    // valuer le dterminant de la matrice.
    double det = this.determinant();
    
    // Vrifier que le dterminant n'est pas NaN.
    if(Double.isNaN(det))
      throw new IllegalArgumentException("Erreur SMatrix3x3 002 : La matrice ne peut pas tre inverse, car le dterminant est NaN (det = " + det + ").");
    
    // Vrifier que le dterminant n'est pas zro.
    if(SMath.nearlyZero(det))
      throw new IllegalArgumentException("Erreur SMatrix3x3 003 : La matrice ne peut pas tre inverse, car le dterminant n'est pas nul (det = " + det + ").");
    
    // valuer la matrice inverse.
    return new SMatrix3x3(
        
        (this.matrix[4]*this.matrix[8] - this.matrix[5]*this.matrix[7])/det , (this.matrix[7]*this.matrix[2] - this.matrix[8]*this.matrix[1])/det , (this.matrix[1]*this.matrix[5] - this.matrix[2]*this.matrix[4])/det ,
        (this.matrix[6]*this.matrix[5] - this.matrix[3]*this.matrix[8])/det , (this.matrix[0]*this.matrix[8] - this.matrix[6]*this.matrix[2])/det , (this.matrix[3]*this.matrix[2] - this.matrix[0]*this.matrix[5])/det ,
        (this.matrix[3]*this.matrix[7] - this.matrix[6]*this.matrix[4])/det , (this.matrix[6]*this.matrix[1] - this.matrix[0]*this.matrix[7])/det , (this.matrix[0]*this.matrix[4] - this.matrix[3]*this.matrix[1])/det
        
        );
              
  }

  /**
   * Mthode pour obtenir la matrice 4x4 correpondant  cette matrice 3x3.
   * Le rsultat aura une 4e linge et colonne etcontenant des 0 sauf  la position m[3,3] = 1.
   * 
   * @return La matrice 4x4 correspondante.
   */
  public SMatrix4x4 toMatrix4x4()
  {
    return new SMatrix4x4(
          this.matrix[0], this.matrix[1], this.matrix[2], 0.0,
          this.matrix[3], this.matrix[4], this.matrix[5], 0.0,
          this.matrix[6], this.matrix[7], this.matrix[8], 0.0,
          0.0, 0.0, 0.0, 1.0
        );
  }
  
  @Override
  public String toString() 
  {
    return "\n"+
           "|" + matrix[0] + "\t" +  matrix[1] + "\t" +  matrix[2] + "|"  + "\n" +  
           "|" + matrix[3] + "\t" +  matrix[4] + "\t" +  matrix[5] + "|"  + "\n" +
           "|" + matrix[6] + "\t" +  matrix[7] + "\t" +  matrix[8] + "|"  + "\n";
  }

  @Override
  public int hashCode() {
    final int prime = 31;
    int result = 1;
    result = prime * result + Arrays.hashCode(matrix);
    return result;
  }

  @Override
  public boolean equals(Object obj) {
    
    if (this == obj)
      return true;
    
    if (obj == null)
      return false;
    
    if (getClass() != obj.getClass())
      return false;
    
    SMatrix3x3 other = (SMatrix3x3) obj;
    
    if (!SArrays.nearlyEquals(matrix, other.matrix))
      return false;
    
    return true;
  }
  
  
}
