/**
 * 
 */
package sim.opengl;

import java.nio.FloatBuffer;
import java.nio.IntBuffer;

import com.jogamp.opengl.GL2;

import sim.graphics.SColor;
import sim.math.SQuaternion;
import sim.math.SVector3d;
import sun.misc.Unsafe;

/**
 * La classe <b>SGL</b> reprsente une classe utilitaire en lien avec des fonctionnalits de la librairie GL de opengl en JOGL.
 * 
 * @author Simon Vezina
 * @since 2016-06-12
 * @version 2023-10-28
 */
public class SGL {

  /**
   * La constante <b>JDK_VERSION_REFERENCE</b> est le numro de rfrence de la version JAVA pour identifier le retrait d'un message d'avertissement (<i>acces warning</i>) tat gal  {@value}.
   */
  public static final int JDK_VERSION_REFERENCE = 13;
  
  /**
   * Mthode pour retirer des <i>warning access</i> (message en rouge) engendrs par la librairie de JOGL.
   */
  public static void removeAccessWarning()
  {
    int jdk_version = Integer.parseInt(System.getProperty("java.specification.version"));
    
    if(jdk_version < JDK_VERSION_REFERENCE)
      removeAccessWarningJDK11andLess();
    else
      System.out.println("Message SGL --- La version de JDK est '" + jdk_version + "' et prsentement, il n'est pas clair que l'on peut y retirer les messages de type WARNING en ROUGE.");      
  }
  
  /**
   * <p>
   * Mthode pour retirer des <i>warning access</i> engendrs par la librairie de JOGL sous JDK-11 et moins.
   * Il est  not que cette mthode n'a pas encore t test pour des versions plus  jour que JDK-11.
   * </p>
   * 
   * <p>
   * Pour le fonctionnement de cette mthode, l'importation de <u>import sun.misc.Unsafe;</u> est requis.
   * </p>
   */
  public static void removeAccessWarningJDK11andLess()
  {
    try {
        
      final java.lang.reflect.Field theUnsafe = Unsafe.class.getDeclaredField("theUnsafe");
      theUnsafe.setAccessible(true);
      final Unsafe unsafe = (Unsafe) theUnsafe.get(null);

      final Class<?> cls = Class.forName("jdk.internal.module.IllegalAccessLogger");
      final java.lang.reflect.Field logger = cls.getDeclaredField("logger");
      unsafe.putObjectVolatile(cls, unsafe.staticFieldOffset(logger), null);
      
      }catch(NoSuchFieldException e) {
        e.printStackTrace();
        throw new RuntimeException(e);
      }catch(IllegalAccessException e) {
        e.printStackTrace();
        throw new RuntimeException(e);
      }catch(ClassNotFoundException e) {
        e.printStackTrace();
        throw new RuntimeException(e);
      }
  }
  
  /**
   * La constante <b>MAX_LIGHT</b> reprsente le nombre maximum de source de lumire supporte.
   */
  public static final int MAX_LIGHT = 8;
  
  /**
   * Mthode permettant de convertir un <b>SVector3d</b> en <b>FloatBuffer</b>.
   * 
   * @param v Le vecteur.
   * @return Le buffer.
   */
  public static FloatBuffer SVector3dToFloatBuffer(SVector3d v)
  {
    float[] tab = { (float)v.getX(), (float)v.getY(), (float)v.getZ(), 0.0f };
      
    return FloatBuffer.wrap(tab);
  }
  
  /**
   * Mthode pour convertir un <b>SColor</b> en <b>FloatBuffer</b>.
   * 
   * @param c La couleur.
   * @return Le buffer.
   */
  public static FloatBuffer SColorToFloatBuffer(SColor c)
  {
    float[] tab = { (float)c.getRed(), (float)c.getGreen(), (float)c.getBlue(), 1.0f };
  
    return FloatBuffer.wrap(tab);
  }
  
  /**
   * ...
   * 
   * @param c
   * @param transparency
   * @return
   */
  public static FloatBuffer SColorToFloatBuffer(SColor c, float transparency)
  {
    float[] tab = { (float)c.getRed(), (float)c.getGreen(), (float)c.getBlue(), 1.0f - transparency };
  
    return FloatBuffer.wrap(tab);
  }
  
  /**
   * ...
   * 
   * @param c
   * @return
   */
  public static float[] SColorToFloatArray(SColor c)
  {
    float[] tab = { (float)c.getRed(), (float)c.getGreen(), (float)c.getBlue() };
  
    return tab;
  }
  
  /**
   * Mthode pour dfinir la couleur des prochains vertexs envoys au <i>pipeline</i>.
   * 
   * @param gl
   * @param color
   */
  public static void setColor(final GL2 gl, SColor color)
  {
    gl.glColor3d(color.getRed(), color.getGreen(), color.getBlue());
  }
  
  /**
   * <p>
   * Mthode pour pousser (<i>gl.glPushMatrix()</i>) une matrice de transformation de type <b>ModelView</b> sur la pile des matrices de transformation de OpenGL.
   * Aprs l'usage de cette matrice, elle peut tre retir avec l'instruction <i>gl.glPopMatrix()</i>.
   * </p>
   * 
   * <p>
   * La matrice de transformation glogale sera construite dans l'ordre suivant :
   * <ul>1) scale</ul>
   * <ul>2) rotation x</ul>
   * <ul>3) rotation y</ul>
   * <ul>4) rotation z</ul>
   * <ul>5) translation</ul>
   * </p>
   * 
   * <p>
   * Il est important de prciser que cette mthode ne prcise par le mode (<i>gl.glMatrix(...)</i>) de la matrice (<i>GL_PROJECTION</i>, <i>GL_MODELVIEW</i>).
   * Ainsi, il est primodial de prciser  OpenGl le mode de la transformation avant de faire l'appel de cette mthode.
   * </p>
   *  
   * @param gl L'objet OpenGL.
   * @param scale La vecteur d'homothtie (<i>scale</i>).
   * @param rotation Le vecteur de rotation dans l'ordre Rx,Ry et Rz.
   * @param translation Le vecteur de translation.
   */
  public static void pushTransformationMatrix(final GL2 gl, SVector3d scale, SVector3d rotation, SVector3d translation)
  {
    // Informer OpenGL qu'il y aura une nouvelle matrice sur la pile
    gl.glPushMatrix();
   
    // Appliquer les matrices dans l'ordre inverse de l'excution.
    // L'ordre d'excution dsir  appliquer par multiplication au vecteur est :
    // 1) scale
    // 2) rotation x
    // 3) rotation y
    // 4) rotation z
    // 5) translation
    
    gl.glTranslated(translation.getX(), translation.getY(), translation.getZ() ); // translation
  
    gl.glRotated(rotation.getZ(), 0.0, 0.0, 1.0); // rotation z
    gl.glRotated(rotation.getY(), 0.0, 1.0, 0.0); // rotation y
    gl.glRotated(rotation.getX(), 1.0, 0.0, 0.0); // rotation x
    
    gl.glScaled(scale.getX(), scale.getY(), scale.getZ());  // scale
  }
  
  /**
   * <p>
   * Mthode pour pousser (<i>gl.glPushMatrix()</i>) une matrice de transformation de type <b>ModelView</b> sur la pile des matrices de transformation de OpenGL.
   * Aprs l'usage de cette matrice, elle peut tre retir avec l'instruction <i>gl.glPopMatrix()</i>.
   * </p>
   * 
   * <p>
   * La matrice de transformation glogale sera construite dans l'ordre suivant :
   * <ul>1) scale</ul>
   * <ul>2) rotation du quaternionx</ul>
   * <ul>3) translation</ul>
   * </p>
   * 
   * <p>
   * Il est important de prciser que cette mthode ne prcise par le mode (<i>gl.glMatrix(...)</i>) de la matrice (<i>GL_PROJECTION</i>, <i>GL_MODELVIEW</i>).
   * Ainsi, il est primodial de prciser  OpenGl le mode de la transformation avant de faire l'appel de cette mthode.
   * </p>
   *  
   * @param gl L'objet OpenGL.
   * @param scale La vecteur d'homothtie (<i>scale</i>).
   * @param rotation Le quaternion de rotation.
   * @param translation Le vecteur de translation.
   */
  public static void pushTransformationMatrix(final GL2 gl, SVector3d scale, SQuaternion q, SVector3d translation)
  {
    // Informer OpenGL qu'il y aura une nouvelle matrice sur la pile
    gl.glPushMatrix();
    
    // Appliquer les matrices dans l'ordre inverse de l'excution.
    // L'ordre d'excution dsir  appliquer par multiplication au vecteur est :
    // 1) scale
    // 2) rotation du quaternion
    // 3) translation
  
    gl.glTranslated(translation.getX(), translation.getY(), translation.getZ() ); // translation
  
    SVector3d axis = q.rotationAxis();
    gl.glRotated(Math.toDegrees(q.rotationAngle()), axis.getX(), axis.getY(), axis.getZ()); // rotation 
      
    gl.glScaled(scale.getX(), scale.getY(), scale.getZ());  // scale
  }
    
  /**
   * Mthode pour effacer une seule texture en mmoire OpenGL.
   * 
   * @param gl L'objet OpenGL.
   * @param id Le numro d'identification ID de la texture en mmoire OpenGL.
   */
  public static void glDeleteTextures(GL2 gl, int id)
  {
    int[] array = { id };
    IntBuffer buffer = IntBuffer.wrap(array);
          
    // Retirer la texture en mmoire OpenGL.
    gl.glDeleteTextures(1, buffer);
  }
  
  /**
   * Mthode pour dfinir les paramtres d'une source de lumire selon l'instruction <i>glLightfv(...)</i> de OpenGL.
   * 
   * @param gl L'objet OpenGL.
   * @param gl_index Le numro d'identification ID de la source de lumire.
   * @param position La position de la source de lumire.
   * @param ambient La couleur ambiante de la source de lumire. 
   * @param diffuse La couleur diffuse de la source de lumire. 
   * @param specular La couleur spculaire de la source de lumire.
   */
  public static void glLightfv(GL2 gl, int gl_index, SVector3d position, SColor ambient, SColor diffuse, SColor specular)
  {
    // Ajustement de la position de la source de lumire
    float[] lightPos = { (float)position.getX(), (float)position.getY(), (float)position.getZ(), 1.0f};
    gl.glLightfv(gl_index, GL2.GL_POSITION, lightPos, 0);
    
    // Ajustement de l'clairage ambiant.
    float[] color_ambient = { (float)ambient.getRed(), (float)ambient.getGreen(), (float)ambient.getBlue(), 1.0f }; 
    gl.glLightfv(gl_index, GL2.GL_AMBIENT,  color_ambient, 0);
    
    // Ajustement de l'clairage diffuse.
    float[] color_diffuse = { (float)diffuse.getRed(), (float)diffuse.getGreen(), (float)diffuse.getBlue(), 1.0f };        
    gl.glLightfv(gl_index, GL2.GL_DIFFUSE,  color_diffuse, 0); 
    
    float[] color_specular = { (float)specular.getRed(), (float)specular.getGreen(), (float)specular.getBlue(), 1.0f };    
    gl.glLightfv(gl_index, GL2.GL_SPECULAR,  color_specular, 0);
  }
}
