/**
 * 
 */
package sim.opengl;

import java.io.File;
import java.io.IOException;
import java.util.HashMap;
import java.util.Map;

import com.jogamp.opengl.GL2;
import com.jogamp.opengl.GLException;
import com.jogamp.opengl.util.texture.Texture;
import com.jogamp.opengl.util.texture.TextureIO;

import sim.util.SBrowser;
import sim.util.SManyFilesFoundException;

/**
 * La classe <b>SGLTextureHandler</b> reprsente un gestionnaire de texture en mmoire OpenGL.
 * 
 * @author Simon Vezina
 * @since 2020-06-25
 * @version 2023-10-28
 */
public class SGLTextureHandler {

  /**
   * La constante <b>NO_TEXTURE_ID</b> reprsente le numro d'identification ID correspondant  aucune texture tant gal  {@value}.
   */
  public final static Integer NO_TEXTURE_ID = -1;
  
  /**
   * Le paramtre <b>opengl_texture_id_map</b> reprsente la carte de correcpondance entre un nom de texture et son numro d'identificaition ID en mmoire OpenGL.
   */
  private final static Map<String, Integer> opengl_texture_id_map = new HashMap<String, Integer>();
  
  /**
   * Mthode pour obtenir le numro d'identification ID de la texture de OpenGL et mmoire.
   * 
   * @param gl L'objet OpenGL.
   * @param file_name Le nom du fichier.
   * @return Le numro ID.
   * @throws SGLException Si le nom est <i>null</i>.
   * @throws SGLException Si le nom du fichier ne correspond pas  une texture pralablement charg en mmoire.
   */
  public static int getTextureID(GL2 gl, String file_name) throws SGLException
  {
    // Cas particulier : Si le nom du fichier est null.
    if(file_name == null)
      throw new SGLException("Erreur SGLTextureHandler 001 : Le nom du fichier est null.");
    
    // Obtenir le ID de la texture si elle a t charg (avec ou sans succs).
    if(opengl_texture_id_map.containsKey(file_name))
      return opengl_texture_id_map.get(file_name).intValue();
    else
      throw new SGLException("Erreur SGLTextureHandler 002 : La texture de nom '" + file_name + "' n'a pas t charge.");
  }
  
  
  /**
   * Mthode pour charger en mmoire une texture en mmoire pour OpenGL et retourne le numro de correspondance ID.
   * 
   * @param gl L'objet OpenGL.
   * @param file_name Le nom du fichier.
   * @return Le numrio ID.
   * @throws SGLException Si le nom du fichier est <i>null</i>.
   * @throws SGLException Si le nom du fichier comporte un problme lors du chargement.
   */
  public static int loadTextureID(GL2 gl, String file_name) throws SGLException
  {
    if(file_name == null)
      throw new SGLException("Erreur SGLTextureHandler 003 : Le nom du fichier est null.");
    
    // Obtenir le ID de la texture si elle a t charg (avec ou sans succs).
    if(opengl_texture_id_map.containsKey(file_name))
      return opengl_texture_id_map.get(file_name).intValue();
    else
    {
      try {
        
        // Chargement de la texture en mmoire pour Opengl et sauvegarde du numro.
        int id = loadTexture(gl, file_name);
        opengl_texture_id_map.put(file_name, id);
        return id;
        
      }catch(SManyFilesFoundException | IOException e) {
        throw new SGLException("Erreur SGLTextureHandler 004 : Le nom du fichier '" + file_name + "' comporte un problme de chargement de texture.", e);
      }catch(GLException e) {
        throw new SGLException("Erreur SGLTextureHandler 005 : Une erreur de type GLException est survenue.", e);
      }
    }  
  }
  
  /**
   * Mthode pour faire la liaison d'une texture OpenGL  partir d'un nom de fichier.
   * 
   * @param gl L'objet OpenGL.
   * @param texture_file_name Le nom de la texture.
   * @throws SGLException Le nom de la texture n'est pas admissible.
   */
  public static void glBindTexture(GL2 gl, String texture_file_name) throws SGLException
  {
    try {
      
      // Accder au numro d'identifiant OpenGL de la texture et la lier dans OpenGL.
      gl.glBindTexture(GL2.GL_TEXTURE_2D, getTextureID(gl, texture_file_name));    
      
    }catch(SGLException e) {
      throw new SGLException("Erreur SGL 001 : La texture '" + texture_file_name + "' n'a pas t li dans OpenGL, car le numro ID n'a pas t gnr.", e);
    }
  }
  
  /**
   * Mthode pour effacer une texture en mmoire OpenGL.
   * 
   * @param gl L'objet OpenGL.
   * @param file_name Le nom de la texture.
   * @return <b>true</b> si la texture  t efface et <b>false</b> sinon.
   */
  public static boolean deleteTexture(GL2 gl, String file_name)
  {
    // Vrifier que la texture est en mmoire.
    if(opengl_texture_id_map.containsKey(file_name))
    {
      // Effacer la texture en mmoire.
      SGL.glDeleteTextures(gl, opengl_texture_id_map.get(file_name).intValue());
      
      // Retire la rfrence  la texture dans la carte.
      opengl_texture_id_map.remove(file_name);
      
      return true;
    }
    else
      return false;
  }
  
  /**
   * Mthode pour effacer l'ensemble des textures en mmoire OpenGL.
   * 
   * @param gl L'objet OpenGL.
   */
  public static void clear(GL2 gl)
  {
    for(String key : opengl_texture_id_map.keySet())
      deleteTexture(gl, key);
  }
  
  /**
   * Mthode pour charger une texture en mmoire.
   * 
   * @param gl La librairie OpenGL 2. 
   * @param file_name Le nom du fichier  charger.
   * @return Le code d'accs OpenGL en mmoire de la texture charge. 
   * @throws SManyFilesFoundException S'il y a eu plusieurs fichiers de trouv avec ce mme nom.
   * @throws IOException Une erreur de type I/O est survenue lors du chargement de la texture.
   * @throws GLException Une erreur de type GL est survenue lors du chargement de la texture.
   */
  private static int loadTexture(final GL2 gl, String file_name) throws SManyFilesFoundException, GLException, IOException
  {
    SBrowser browser = new SBrowser();
    File file = browser.findFile(file_name);
    Texture texture = TextureIO.newTexture(file, true);
        
    return texture.getTextureObject(gl);
  }
  
}
