/**
 * 
 */
package sim.opengl;

import com.jogamp.opengl.GL2;
import com.jogamp.opengl.glu.GLU;

import sim.graphics.camera.SCamera;
import sim.math.SVector3d;

/**
 * La classe <b>SGLUtil</b> reprsente une classe utilitaire en lien avec des fonctionnalits de la librairie GLU de opengl en JOGL.
 * 
 * @author Simon Vzina
 * @since 2016-06-12
 * @version 2023-10-25
 */
public class SGLU {

  /**
   * La constante <b>MAX_RATIO_ZFAR_DIVIDE_ZNEAR</b> correspond au ratio maximum autoris pour Z_far / Z_near tant gal  {@value}.
   * Ce choix est prsentement dtermin sur une exprience personnelle d'un nombre limit de scne. 
   * Cette valeur est sujet  rvaluation (2023-10-15). 
   */
  public static double MAX_RATIO_ZFAR_DIVIDE_ZNEAR = 1e6;
  	
  /**
   * <p>
   * Mthode pour gnrer  l'aide d'opengl la matrice de transformation associ au positionnement de la camra et son orientation.
   * </p>
   * 
   * <p>
   * Il est important de prciser que cette mthode ne prcise pas le mode de la matrice (<i>GL_MODELVIEW</i>).
   * Il est alors important de dfinir le mode avant de faire l'appel de cette mthode (ex : <i>gl.glMatrixMode(GL2.GL_MODELVIEW)</i>).
   * </p>
   * 
   * @param position La position de la camera.
   * @param look_at La position o regarde la camera.
   * @param up L'orientation haut de la camera.
   */
  public static void gluLookAt(SVector3d position, SVector3d look_at, SVector3d up)
  {
    // Charger la version des fonctionnalits de GLU
    final GLU glu = new GLU();
    
    // Utiliser les fonctionnalits de la librairie GLU pour faire la transformation de l'orientation de la camra
    glu.gluLookAt(position.getX(), position.getY(), position.getZ(),
                  look_at.getX(), look_at.getY(), look_at.getZ(),
                  up.getX(), up.getY(), up.getZ());   
  }
	  
  /**
   * <p>
   * Mthode pour gnrer  l'aide d'opengl la matrice de transformation associ au positionnement de la camra et son orientation.
   * </p>
   * 
   * <p>
   * Il est important de prciser que cette mthode ne prcise pas le mode de la matrice (<i>GL_MODELVIEW</i>).
   * Il est alors important de dfinir le mode avant de faire l'appel de cette mthode (ex : <i>gl.glMatrixMode(GL2.GL_MODELVIEW)</i>).
   * </p>
   * 
   * @see https://www.khronos.org/opengl/wiki/GluLookAt_code
   * @see https://www.dei.isep.ipp.pt/~matos/cg/docs/manual/gluLookAt.3G.html
   * @param gl
   * @param camera
   */
  public static void gluLookAt(final GL2 gl, SCamera camera)
  {
    SVector3d forward = camera.getFront();
    SVector3d right = camera.getRight();
    
    // Important : up soit perpendiculaire  forward (dj le cas dans SCamera).
    SVector3d up = camera.getUp();
    
    SVector3d position = camera.getPosition();
        
    //---------------------------------------------------------------------------------------------------------------------------------------------
    // Cration d'une matrice particulire : (https://www.dei.isep.ipp.pt/~matos/cg/docs/manual/gluLookAt.3G.html)
    // The matrix maps the reference point to the negative z axis and the eye point to the origin. 
    // When a typical projection matrix is used, the center of the scene therefore maps to the center of the viewport. 
    // Similarly, the direction described by the UP vector projected onto the viewing plane is mapped to the positive y axis 
    // so that it points upward in the viewport. The UP vector must not be parallel to the line of sight from the eye point to the reference point.
    //---------------------------------------------------------------------------------------------------------------------------------------------
    //  M=( s[0]    s[1]    s[2]   0
    //      u[0]    u[1]    u[2]   0
    //     -f[0]   -f[1]   -f[2]   0
    //        0       0       0    1)
    //
    // Remarque : C'est la version transpose que sera  multiplit dans le systme OpenGL.
    double[] matrixTranspose = {
         right.getX(),  up.getX(), -forward.getX(),  0.0,
         right.getY(),  up.getY(), -forward.getY(),  0.0,
         right.getZ(),  up.getZ(), -forward.getZ(),  0.0,
                  0.0,        0.0,             0.0,  1.0
    };
        
    // Multiplier dans la matrice de transformation particulire.
    gl.glMultMatrixd(matrixTranspose, 0);
    
    // Multiplier par la matrice de translation en lien avec la position de la camera.
    gl.glTranslated(-position.getX(), -position.getY(), -position.getZ());
  }
  
  /**
   * <p>
   * Mthode pour gnrer  l'aide d'opengl la matrice de transformation associ au positionnement de la camra et son orientation.
   * </p>
   * 
   * <p>
   * Il est important de prciser que cette mthode ne prcise pas le mode de la matrice (<i>GL_MODELVIEW</i>).
   * Il est alors important de dfinir le mode avant de faire l'appel de cette mthode (ex : <i>gl.glMatrixMode(GL2.GL_MODELVIEW)</i>).
   * </p>
   *  
   * @param camera La camera.
   */
  public static void gluLookAt(SCamera camera)
  {
    gluLookAt(camera.getPosition(), camera.getLookAt(), camera.getUp());
  }
   
  /**
   * Mthode pour pousser la matrice de la pyramide de vue (<i>viewing frustum</i>) afin qu'elle correspond aux coordonnes monde du systme.
   * Il est important d'avoir pralablement bien dfini le bon mode des matrice (ex : <i>gl.glMatrixMode(GL2.GL_PROJECTION)</i>) et
   * d'avoir pouss une matrice identit prcdemment (<i>gl.glLoadIdentity()</i>).
   *    
   * <p>
   * AUTRE DOCUMENTATION : Depth buffer precision is affected by the values specified for zNear and zFar. 
   * The greater the ratio of zFar to zNear is, the less effective the depth buffer will be at distinguishing between surfaces that are near each other. 
   * If
   * <ul><i>r</i> = zFar / znear</ul>
   * roughly log2(<i>r</i>) bits of depth buffer precision are lost. Because <i>r</i> approaches infinity as zNear approaches 0, zNear must never be set to 0.
   * </p>
   * 
   * @param gl L'objet OpenGL.
   * @param fovyInDegrees L'angle d'ouverture de la camra (en degrs).
   * @param aspectRatio Le ratio largeur sur hauteur de la fentre de vue (<i>width</i> / <i>height</i>);
   * @param znear La distance camra et cran avant de la pyramide de vue (<i>front clipping plane</i>).
   * @param zfar La distance camra  l'cran arrire de la pyramide de vue (<i>back clipping plane</i>).
   * 
   * @see https://www.khronos.org/opengl/wiki/GluPerspective_code
   * @see https://registry.khronos.org/OpenGL-Refpages/gl2.1/xhtml/gluPerspective.xml
   */
  public static void gluPerspective(final GL2 gl, double fovyInDegrees, double aspectRatio, double znear, double zfar) throws SGLException
  {
    if(znear == 0.0)
      throw new SGLException("Erreur SGLU 001 : Le znear ne peut pas tre 0.0.");
    
    if(zfar < znear)
      throw new SGLException("Erreur SGLU 002 : Le znear = " + znear + " doit tre plus petit que zfar = " + zfar + ".");
    
    if(fovyInDegrees < 0.0 || aspectRatio < 0.0)
      throw new SGLException("Erreur SGLU 003 : Les paramtres fovyInDegrees et aspectRatio doit tre positive.");
    
    if(zfar/znear > MAX_RATIO_ZFAR_DIVIDE_ZNEAR)
      throw new SGLException("Erreur SGLU 004 : Le rapport zfar/znear = " + zfar/znear + " est suprieur  " + MAX_RATIO_ZFAR_DIVIDE_ZNEAR + " ce qui ne permet pas au z-buffer de bien identifier la visibilit des objets  grande distance. Comme solution, vous pouvez loigner znear ou rapprocher zfar.");
        
    // Dfinition de f : f = cotan(fovy/2.0)
    double f = 1.0 / Math.tan(Math.toRadians(fovyInDegrees/2.0));
    double znear_minus_zfar = znear-zfar;
    
    // Remarque : C'est la version transpose que sera  multiplit dans le systme OpenGL.
    double[] matrix = {
        f / aspectRatio, 0.0,                               0.0,  0.0,
                    0.0,   f,                               0.0,  0.0,
                    0.0, 0.0,   (zfar+znear)/(znear_minus_zfar), -1.0,
                    0.0, 0.0, 2.0*zfar*znear/(znear_minus_zfar),  0.0
    };
    
    // Multiplication de cette matrice avec celle dj en place (habituellement, l'identit).
    gl.glMultMatrixd(matrix, 0);
  }
	
}
