/**
 * 
 */
package sim.opengl;

import java.util.function.DoubleConsumer;

import sim.util.SChronometer;

/**
 * La classe <b>SSimulationThread</b> est un objet reprsentant une thread ralisant des simulations.
 *  
 * @author Simon Vezina
 * @since 2021-08-15
 * @version 2023-10-12
 */
public class SSimulationThread extends Thread {

  /**
   * ...
   */
  private static boolean DEFAULT_PLAY = true;
  
  /**
   * La constante <b>MINIMUM_SLEEP</b> reprsente le temps d'arrt minimale dans la thread  effectuer peut importe le temps de calcul requis.
   */
  private static final long MINIMUM_SLEEP = 5;
  
  /**
   * ...
   */
  private static final int DEFAULT_NB_STEP = 1;
  
  /**
   * ...
   */
  static final double DEFAULT_TIME_STEP = 0.01;
  
  
  /**
   * ...
   */
  final DoubleConsumer simulation_function;
  
  /**
   * ...
   */
  protected int nb_step;
  
  /**
   * ...
   */
  protected double time_step;
  
  /**
   * ...
   */
  protected double simulation_time;
  
  /**
   * La variable <b>accumulated_real_time</b> reprsente le temps rel accumul depuis depuis la dernire mise en marche de la simulation.
   */
  protected double accumulated_real_time;
  
  /**
   * ...
   */
  protected SChronometer real_time_chrono;
  
  /**
   * ...
   */
  protected double step_time_required;
  
  /**
   * ...
   */
  protected boolean is_simulation_finished;
  
  /**
   * La variable <b>play</b> dtermine s'il y aura volution de la simulation. 
   * Lorsque ce paramtre sera <i>false</i>, la simulation sera stopp.
   */
  protected boolean play;
  
  /**
   * Constructeur du simulateur.
   * 
   * @param function La fonction qui va dfinir le contenu de la simulaion excut par le thread.
   */
  public SSimulationThread(DoubleConsumer function)
  {
    this.play = DEFAULT_PLAY;
    
    this.simulation_function = function;
    
    this.nb_step = DEFAULT_NB_STEP;
    this.time_step = DEFAULT_TIME_STEP;
       
    this.simulation_time = 0.0;
    this.accumulated_real_time = 0.0;
    
    this.real_time_chrono = new SChronometer();
    
    this.step_time_required = Double.POSITIVE_INFINITY;
    
    this.is_simulation_finished = false;
  }

  /**
   * Mthode pour obtenir le nombre d'itration du time_step qui sera ralis par cycle de simulation.
   * 
   * @return Le nombre de time_step.
   */
  public int getNbStep()
  {
    return this.nb_step;
  }
  
  /**
   * Mthode pour obtenir le temps de simulation pour chaque tape d'un cycle de simulation (il peut y avoir plusieurs tape dans un cycle).
   * 
   * @return Le temps de simulation d'une tape.
   */
  public double getTimeStep()
  {
    return this.time_step;
  }
  
  /**
   * ...
   * 
   * @return
   */
  public double getSimulationTime()
  {
    return this.simulation_time;
  }
  
  /**
   * Mthode pour obtenir le temps rel coul depuis le dbut du calcul de la simulation.
   * 
   * @return Le temps rel coul en seconde.
   */
  public double getRealTime()
  {
    if(this.real_time_chrono.asStart())
      return this.accumulated_real_time + this.real_time_chrono.getTime();
    else
      return this.accumulated_real_time;
  }
    
  /**
   * Mthode pour dfinir le temps de simulation par itration dans un cycle.
   * 
   * @param time Le temps de simulation.
   * @throws Si time est infrieur  0.
   */
  public void setTimeStep(double time) throws IllegalArgumentException
  {
    if(time < 0.0)
      throw new IllegalArgumentException("Erreur SSimulationThread 001 : Le temps doit tre suprieur  0 (time = " + time + ").");
    
    this.time_step = time;
  }
  
  /**
   * Mthode pour dfinir le nombre d'tape dans un cycle de simulation.
   * 
   * @param nb Le nombre d'tape.
   * @throws Si nb est infrieur  1.
   */
  public void setNbStep(int nb) throws IllegalArgumentException
  {
    if(nb < 1)
      throw new IllegalArgumentException("Erreur SSimulationThread 001 : Le nombre de step doit tre suprieur  1 (nb = " + nb + ").");
    
    this.nb_step = nb;
  }
  
  /**
   * Mthode pour dterminer si le simulateur actif. Dans ce cas, les calculs de la simulation son en action.
   * 
   * @return <b>true</b> si le simulateur est actif et <b>false</b> sinon.
   */
  public boolean isPlaying()
  {
    return this.play;
  }
  
  /**
   * Mthode pour dterminer si le simulateur est en pause. Dans ce cas, les calculs de la simulation son interrompu. 
   * 
   * @return <b>true</b> si le simulateur est en pause et <b>false</b> sinon.
   */
  public boolean isPause()
  {
    return !isPlaying();
  }
  /**
   * ...
   * 
   * @param play
   */
  public void setPlaying(boolean play)
  {
    // Ajustement du chronomtre qui mesure de l'coulement du temps rel pour caluler la simulation.
    if(play)
      this.real_time_chrono.start();
    else
      if(this.real_time_chrono.asStart())
      {
        this.accumulated_real_time += this.real_time_chrono.getTime();
        this.real_time_chrono.reset();
      }
        
    // Dfinir l'tat de la simulation.
    this.play = play;
  }
  
  /**
   * Mthode pour inverser le mode de la simulation (action <--> arrt).
   */
  public void switchPlaying()
  {
    if(isPlaying())
      setPlaying(false);
    else
      setPlaying(true);
  }
  
  /**
   * Mthode pour terminer la simulation. L'usage de cette mthode forcera la fin de la boucle de la mthode run().
   */
  public void endSimulation()
  {
    is_simulation_finished = true;
  }
  
  /**
   * <p>
   * Mthode pour rinitialiser le temps de la simulation.
   * </p>
   * <p>
   * Lors de l'excution de cette mthode, le simulateur se met automatiquement en pause.
   * </p>
   */
  public void reset()
  {
    this.simulation_time = 0.0;
    this.accumulated_real_time = 0.0;
    this.real_time_chrono.reset();
    
    setPlaying(false);
  }
  
  @Override
  public void run()
  {
    try {
      
      while(!is_simulation_finished) 
      {
        // Outil pour mesurer le temps requis pour raliser la simulation.
        SChronometer chrono = new SChronometer();
        chrono.start();
        
        //--------------------------------------------------
        // Raliser la simulation si le mode play est activ
        //--------------------------------------------------
        if(this.play)
        {
          
          // Raliser le nombre d'tape d'coulement de temps.
          for(int i = 0; i < nb_step; i++)
            simulation_function.accept(this.time_step);             // FONCTION  DFINIR DANS LE THREAD (double Consumer).
        
          chrono.stop();
        
          // Enregistrer le temps requis pour raliser la simulation.
          step_time_required = (double) (chrono.getTime());
        
          // Modifier le temps de simulation.
          this.simulation_time += this.time_step * nb_step;
        }
        
        // Raliser la pause.
        Thread.sleep(MINIMUM_SLEEP);  
      }
    } catch(InterruptedException e) {
      e.printStackTrace();
    } 
  }
  
}
