/**
 * 
 */
package sim.physics;

import sim.exception.SNoImplementationException;
import sim.math.SVector3d;

/**
 * La classe <b>SNumericalIntegrationMotion</b> permet de ralise l'intgration numrique des quations de la cinmatique.
 * 
 * @author Simon Vezina
 * @since 2020-05-25
 * @version 2023-10-30 (Version : Les orbites (2023-10-30))
 */
public class SNumericalIntegrationMotion {

  /**
   * La constante <b>EULER</b> reprsente le numro d'identification de l'algorithme numrique de <i>Euler</i> tant gal  {@value}.
   */
  public final static int EULER = 0;
  
  /**
   * La constante <b>EULER_SECOND_ORDER</b> reprsente le numro d'identification de l'algorithme numrique de <i>Euler de 2e ordre</i> tant gal  {@value}.
   */
  public final static int EULER_SECOND_ORDER = 1;
  
  /**
   * La constante <b>RK4</b> reprsente le numro d'identification de l'algorithme numrique de <i>RK4 : Runge-Kutta d'ordre 4</i> tant gal  {@value}.
   */
  public final static int RK4 = 2;
  
  /**
   * Mthode pour calculer la position et la vitesse prdites pour une particule subissant une force rsultante F durant un coulement de temps dt.
   * 
   * @param p La particule.
   * @param F La force rsultante (N).
   * @param dt L'coulement de temps (s).
   * @param numerical_algorithm Le numro d'identification de l'algorithme numrique utilis.
   * @return Les composants cintiques prdits sur la particule.
   * @throws IllegalArgumentException Si le numro de l'algorithme n'est pas reconnu.
   */
  public static SKineticComponent computePredictedComponent(SParticle p, SApplyForce F, double dt, int numerical_algorithm) throws IllegalArgumentException
  {
    switch(numerical_algorithm)
    {
      case EULER :              return eulerForceIntegration(p, F, dt);
      case EULER_SECOND_ORDER : return eulerSecondOrderForceIntegration(p, F, dt);
      case RK4 :                return RK4ForceIntegration(p, F, dt);
      
      default : throw new IllegalArgumentException("Erreur SNumericalIntegrationMotion 001 : Le numro d'algorithme '" + numerical_algorithm + "' n'est pas reconnu.");
    }
  }
  
  /**
   * Mthode pour calculer la position et la vitesse prdites pour une particule subissant une force rsultante F durant un coulement de temps dt
   * selon l'algorithme numrique de <i>Euler</i>.
   * 
   * @param p La particule.
   * @param F La force rsultante (N).
   * @param dt L'coulement de temps (s).
   * @return Les composants cintiques prdits sur la particule.
   */
  public static SKineticComponent eulerForceIntegration(SParticle p, SApplyForce F, double dt)
  {
    throw new SNoImplementationException();
    
  }
  
  /**
   * Mthode pour calculer la position et la vitesse prdites pour une particule subissant une force rsultante F durant un coulement de temps dt
   * selon l'algorithme numrique de <i>Euler de 2e ordre</i>.
   * 
   * @param p La particule.
   * @param F La force rsultante (N).
   * @param dt L'coulement de temps (s).
   * @return Les composants cintiques prdits sur la particule.
   */
  public static SKineticComponent eulerSecondOrderForceIntegration(SParticle p, SApplyForce F, double dt)
  {
    throw new SNoImplementationException();
   
  }
  
  /**
   * Mthode implmentant l'algorithme de Euler pour le 1er ordre pour la position
   * afin de dterminer les composants cinmatiques d'une particule.
   * 
   * @param c_i Les composants cintiques initiaux du mouvement (position et vitesse).
   * @param a_i L'acclration.
   * @param dt L'coulement du temps.
   * @return Les composants cintiques finaux du mouvement (position et vitesse).
   */
  public static SKineticComponent euler(SKineticComponent c_i, SVector3d a_i, double dt)
  {
    throw new SNoImplementationException();
    
  }
  
  /**
   * Mthode implmentant l'algorithme de Euler pour le 2e ordre pour la position
   * afin de dterminer les composants cinmatiques d'une particule.
   * 
   * @param c_i Les composants cintiques initiaux du mouvement (position et vitesse).
   * @param a_i L'acclration.
   * @param dt L'coulement du temps.
   * @return Les composants cintiques finaux du mouvement (position et vitesse).
   */
  public static SKineticComponent eulerSecondOrder(SKineticComponent c_i, SVector3d a_i, double dt)
  {
    throw new SNoImplementationException();
    
  }
  
  /**
   * Mthode implmentant l'algorithme du RK4 (Mthode de Runge-Kutta d'ordre 4) pour la position.
   * afin de dterminer les composants cinmatiques d'une particule.
   * 
   * @param p La particule.
   * @param F Le calcultateur de force.
   * @param dt L'coulement du temps.
   * @return Les composants cintiques finaux du mouvement (position et vitesse).
   */
  public static SKineticComponent RK4ForceIntegration(SParticle p, SApplyForce F, double dt)
  {
    throw new SNoImplementationException();
    
  }
  
  /**
   * Mthode permettant d'obtenir la 1re position et vitesse finale de l'algorithme du RK4.
   * 
   * @param c_i Les composants cintiques initiaux du mouvement (position et vitesse).
   * @param a_i L'acclration initiale.
   * @param dt L'coulement de temps.
   * @return Les composants cintiques finaux.
   */
  protected static SKineticComponent firstFinalStateRK4(SKineticComponent c_i, SVector3d a_i, double dt)
  {
    throw new SNoImplementationException();
   
  }
  
  /**
   * Mthode permettant d'obtenir la position et la vitesse de demi-temps de l'algorithme du RK4.
   * 
   * @param c_i Les composants cintiques initiaux du mouvement (position et vitesse).
   * @param a_i L'acclration initiale.
   * @param a_f1 L'acclration de la 1re position et vitesse finale.
   * @param dt L'coulement de temps.
   * @return Les composants cintiques finaux.
   */
  protected static SKineticComponent middleTimeStateRK4(SKineticComponent c_i, SVector3d a_i, SVector3d a_f1, double dt)
  {
    throw new SNoImplementationException();
    
  }
  
  /**
   * Mthode permettant d'obtenir la 2e position et vitesse finale de l'algorithme du RK4.
   * 
   * @param c_i Les composants cintiques initiaux du mouvement (position et vitesse).
   * @param a_i L'acclration initiale.
   * @param a_f1 L'acclration de la 1re position et vitesse finale.
   * @param dt L'coulement de temps.
   * @return Les composants cintiques finaux.
   */
  protected static SKineticComponent secondFinalStateRK4(SKineticComponent c_i, SVector3d a_i, SVector3d a_f1, double dt)
  {
    throw new SNoImplementationException();
    
  }
  
  /**
   * Mthode permettant d'obtenir l'tat final de la position et la vitesse de l'algorithme du RK4.
   * 
   * @param c_i Les composants cintiques initiaux du mouvement (position et vitesse).
   * @param a_i L'acclration initiale.
   * @param a_mid L'acclration de position et vitesse de demi-temps.
   * @param a_f2 L'acclration de la 2e position et vitesse finale.
   * @param dt L'coulement du temps.
   * @return Les composants cintiques finaux de l'algorithme du RK4.
   */
  protected static SKineticComponent finalStateRK4(SKineticComponent c_i, SVector3d a_i, SVector3d a_mid, SVector3d a_f2, double dt)
  {
    throw new SNoImplementationException();
    
  }
  
}
