/**
 * 
 */
package sim.physics;

import java.util.ArrayList;
import java.util.List;

import sim.exception.SConstructorException;
import sim.math.SVector3d;
import sim.readwrite.object.SParticleRW;
import sim.readwrite.object.SParticleSystemRW;
import sim.util.SReadingException;


/**
 * La classe <b>SParticleSystem</b> reprsente un systme de particle.
 * 
 * @author Simon Vzina
 * @since 2017-05-31
 * @version 2023-10-24
 */
abstract public class SParticleSystem implements SApplyForce {

  /**
   * La constante <b>DEFAULT_MASS</b> reprsente la masse par dfaut d'une particule.
   */
  protected static final double DEFAULT_MASS = 1.0;
  
  /**
   * La constante <b>DEFAULT_ELECTRIC_CHARGE</b> reprsente la charge lectrique par dfaut d'une particule.
   */
  protected static final double DEFAULT_ELECTRIC_CHARGE = 1.0;
  
  /**
   * La constante <b>DEFAULT_SPEED</b> reprsente la vitesse par dfaut d'une particule.
   */
  protected static final SVector3d DEFAULT_SPEED = SVector3d.ZERO;
   
  /**
   * La variable <b>particlesLi</b> reprsente la liste des particules du systme.
   */
  protected final List<SParticle> particles_list;
  
  //----------------
  // CONSTRUCTEUR //
  //----------------
  
  /**
   * Constructeur d'un systme de particules  partir d'un tableau de particules.
   * 
   * @param particles La tableau des particles.
   */
  public SParticleSystem(SParticle ... particles)
  {
    // Crer la liste interne des particules du systme. 
    this.particles_list = new ArrayList<SParticle>();
    
    //particles_reading_list = new ArrayList<SParticleRW>();
    
    // Remplir la liste des particules.
    for(SParticle p : particles)
      this.particles_list.add(p);
  }
  
  /**
   * Constructeur d'un systme de particules  partir d'une liste de particules.
   * 
   * @param list La liste des particules.
   */
  public SParticleSystem(List<SParticle> list)
  {
    // Crer la liste interne des particules du systme. 
    this.particles_list = new ArrayList<SParticle>();
      
    // Remplir la liste des particules.
    for(SParticle p : list)
      this.particles_list.add(p);
  }
  
  
  /**
   * Constructeur d'un systme de particule  partir de la lecture d'un fichier txt dcrivant une lise de particule.
   * 
   * @param file_name Le nom du fichier.
   * @throws SConstructorException S'il y a eu une erreur le de la construction de l'objet.
   */
  public SParticleSystem(String file_name) throws SConstructorException
  {
    // Crer la liste interne des particules du systme. 
    this.particles_list = new ArrayList<SParticle>();
    
    // Lecture du fichier,
    SParticleSystemRW reader = new SParticleSystemRW(file_name);
    
    try {
      reader.read();
    }catch(SReadingException e) {
      throw new SConstructorException("Erreur SParticulesSystem 001 : Une erreur est survenue lors de la lecture du fichier '" + file_name +"'.", e);
    }
    
    // Construire les particules pour la liste du systeme
    List<SParticleRW> list = reader.getParticleRWList();
    
    for(SParticleRW p : list)
      this.particles_list.add(p.toSParticle());
  }
  
  //-----------
  // MTHODE //
  //-----------
  
  /**
   * Mthode pour obtenir le nombre de particules dans le systme.
   * 
   * @return Le nombre de particules dans le systme.
   */
  public int getNbParticles()
  {
    return this.particles_list.size();
  }
  
  /**
   * Mthode pour ajouter une particule au systme. Cette particule aura ajoute  la fin de la liste des particule du systme (index maximal).
   * 
   * @param p La particule ajoute au systme.
   */
  public void add(SParticle p)
  {
    this.particles_list.add(p);
  }
  
  /**
   * Mthode pour obtenir une particule du systme  partir de son numro d'indice dans la liste des particules appartenant au systme.
   * 
   * @param index L'indice de la particule.
   * @return La particule.
   * @throws IndexOutOfBoundsException Si l'indice n'est pas admissible.
   */
  public SParticle getParticle(int index) throws IndexOutOfBoundsException 
  {
    return this.particles_list.get(index);
  }
  
  /**
   * Mthode pour retirer la dernire particule insere dans le systme.
   * 
   * @return La dernire particule insere dans le systme et qui fut retire.
   * @throws UnsupportedOperationException Si le systme ne possde pas de particule.
   */
  public SParticle removeLast() throws UnsupportedOperationException
  {
    // Vrifier que la liste n'est pas vide.
    if(this.particles_list.size() == 0)
      throw new UnsupportedOperationException("Erreur SParticlesSystem 003 : La liste de particule est vide.");
      
    return this.particles_list.remove(particles_list.size()-1);
  }
  
  /**
   * Mthode pour retirer l'ensemble des particules du systme.
   */
  public void clear()
  {
    this.particles_list.clear();
  }
  
  /**
   * Mthode pour dplacer l'ensemble des particules du systme sur un intervalle de temps selon les lois de la dynamique de Newton (2e loi de Newton).
   * 
   * @param dt L'intervalle de temps.
   */
  public void move(double dt, int numerical_algorithm)
  {
    // Calculer la position et vitesse prdites pour l'ensemble des particules.
    for(SParticle p : this.particles_list)
      p.setKineticPredictedComponent(SNumericalIntegrationMotion.computePredictedComponent(p, this, dt, numerical_algorithm));
    
    // Dfinir la position et vitesse prdites comme tant les composants mis  jour.
    for(SParticle p : this.particles_list)
      p.switchKineticComponent();
  }
    
  @Override
  public SVector3d force(SParticle particle)
  {
    // Conteneur de la somme des forces.
    SVector3d force_sum = SVector3d.ZERO;
    
    // Itrer sur l'ensemble des charges du systme.
    for(SParticle p : particles_list)
    {
      // Ne pas appliquer la force avec une particule de mme numro d'identification.
      if(!p.asSameID(particle))
        force_sum = force_sum.add(this.interactionForce(p, particle));
    }
    return force_sum;
  }
  
  /**
   * Mthode pour obtenir la force applique sur une particule (d'indice i) par les autres particules du systme.
   * 
   * @param index L'indice de la particule qui va subir la force.
   * @return La force applique sur la particule.
   * @throws IndexOutOfBoundsException Si l'indice ne concorde pas avec une particule du systme.
   */
  public SVector3d force(int index) throws IndexOutOfBoundsException
  {
    return force(this.particles_list.get(index));
  }
       
  /**
   * Mthode pour obtenir la quantit de mouvement (<i>momentum</i>) du systme (de l'ensemble des particules).
   * 
   * @return La quantit de mouvement totale.
   */
  public SVector3d momentum()
  {
    SVector3d momentum_sum = SVector3d.ZERO;
    
    // Itrer sur l'ensemble des charges du systme.
    for(SParticle p : particles_list)
      momentum_sum = momentum_sum.add(SMechanics.momentum(p.getMass(), p.getSpeed()));
    
    return momentum_sum;
  }
  
  /**
   * Mthode pour obtenir l'nergie cintique (<kinetic energy</i>) du systme (de l'ensemble des particules).
   * 
   * @return L'nergie cintique totale.
   */
  public double kineticEnergy()
  {
    double K = 0.0;
    
    // Itrer sur l'ensemble des charges du systme.
    for(SParticle p : particles_list)
      K += SMechanics.kineticEnergy(p.getMass(), p.getSpeed());
    
    return K;
  }
  
  /**
   * Mthode pour obtenir la moment cintique (<i>angular momentum</i>) du systme (de l'ensemble des particules).
   * Ce moment cintique sera valu par rapport  l'origine du sytme d'axe positionnant l'ensemble des particules.
   * 
   * @return Le moment cintique total.
   */
  public SVector3d angularMomentum()
  {
    SVector3d L = SVector3d.ZERO;
    
    // Itrer sur l'ensemble des charges du systme.
    for(SParticle p : particles_list)
      L = L.add(SMechanics.angularMomentum(p.getMass(), p.getPosition(), p.getSpeed()));
    
    return L;
  }
  
  /**
   * <p>
   * Mthode pour obtenir l'nergie potentielle du systme (de l'ensemble des particules).
   * </p>
   * <p>
   * Le type d'nergie potentielle value dpendra de la classe qui va redfinir la mthode <i>interactionEnergy(...)</i>.
   * </p>
   * 
   * @return L'nergie potentielle totale.
   */
  public double potentialEnergy()
  {
    double U = 0.0;
    
    SParticle p1;
    SParticle p2;
    
    // Itrer sur l'ensemble des charges du systme.
    for(int i = 0; i < particles_list.size(); i++)
      for(int j = i + 1; j < particles_list.size(); j++)
      {
        p1 = this.particles_list.get(i);
        p2 = this.particles_list.get(j);
      
        U += this.interactionEnergy(p1, p2);
      }
    
    return U;
  }
  
  /**
   * Mthode pour obtenir l'nergie totale du systme (de l'ensemble des particules).
   * Cette nergie comprend l'nergie cintique et l'nergie potentielle.
   * 
   * @return L'nergie totale.
   */
  public double totalEnergy()
  {
    return this.potentialEnergy() + this.kineticEnergy();
  }
    
  /**
   * Mthope pour obtenir la force applique sur une particule (cible = <i>target</i>) par une autre particule (source).
   * Cette mthode value donc l'interaction entre les deux particules, mais le vecteur est valu pour s'appliquer sur la cible.
   * 
   * @param p_source La particule qui applique la force.
   * @param p_target La particule qui subit la force.
   * @return La force applique par la source sur la cible.
   * @throws IllegalArgumentException Si les deux particules ont le mme numro d'identification.
   */
  abstract public SVector3d interactionForce(SParticle p_source, SParticle p_target) throws IllegalArgumentException;
  
  /**
   * Mthode pour obtenir l'nergie potentielle entre deux particules.
   * 
   * @param p1 La 1re particule.
   * @param p2 La 2e particule.
   * @return L'nergie potentielle.
   * @throws IllegalArgumentException Si les deux particules sont identiques, car ceux-ci se doivent d'tre diffrentes.
   */
  abstract public double interactionEnergy(SParticle p1, SParticle p2) throws IllegalArgumentException;
  
  @Override
  public String toString() 
  {
    StringBuffer bString = new StringBuffer();
    
    for(int i = 0; i < particles_list.size(); i++)
    {
      bString.append((i+1) + "- ").append(this.particles_list.get(i).toString()).append("\n");
    }
    
    return bString.toString();
  }

}
