/**
 * 
 */
package sim.readwrite;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Locale;
import java.util.StringTokenizer;

import sim.exception.SConstructorException;
import sim.exception.SRuntimeException;
import sim.graphics.SColor;
import sim.math.SMath;
import sim.math.SVector3d;
import sim.math.SVectorUV;
import sim.util.SBrowser;
import sim.util.SBufferedReader;
import sim.util.SInitializationException;
import sim.util.SKeyWordDecoder;
import sim.util.SLog;
import sim.util.SManyFilesFoundException;
import sim.util.SReadingException;
import sim.util.SStringUtil;

/**
 * La classe abstraite <b>SAbstractRW</b> reprsente un objet pouvant tre initialis par une lecture dans un fichier txt. 
 * 
 * @author Simon Vzina
 * @since 2015-01-10
 * @version 2023-10-02
 */
public abstract class SAbstractRW implements SRW {

  //--------------
  // CONSTANTES //
  //--------------
  
  /**
   * La constante <b>KEYWORD_PARAMETER</b> correspond  un tableau contenant l'ensemble des mots cls 
   *  utiliser reconnus lors de la dfinition de l'objet par une lecture en fichier.
   */
  private static final String[] KEYWORD_PARAMETER = { 
      SKeyWordDecoder.KW_END, 
      SKeyWordDecoder.KW_FILE_NAME, 
      SKeyWordDecoder.KW_COMMENT 
      };
  
  /**
   * La constante <b>DEFAULT_FILE_NAME</b> reprsente le nom du fichier par dfaut tait {@value}.
   */
  protected static final String DEFAULT_FILE_NAME = null;
  
  /**
   * La constante <b>TRUE_EXPRESSION</b> corresond  la liste des mots reconnus par le systme comme tant une expression <b>vrai</b>.
   */
  public static final String[] TRUE_EXPRESSION = { "true", "vrai", "yes", "oui", "on" };   
  
  /**
   * La constante <b>FALSE_EXPRESSION</b> correspond  la liste des mots reconnus par le systme comme tant une expression <b>fausse</b>.
   */
  public static final String[] FALSE_EXPRESSION = { "false", "faux", "no", "non", "off" }; 
    
  /**
   * La constante <b>COMMENT_LETTER</b> correspond aux premires lettres d'une ligne lues qui seront considres comme une ligne de commentaire.
   */
  private static final String COMMENT_LETTER = "#/*-!@";
  
  /**
   * La constante <b>OPEN_COMMENT_MODE</b> correspond au mot cl dsignant l'ouveture d'un bloc de commentaire.
   */
  private static final String OPEN_COMMENT_MODE = "/*";
  
  /**
   * La constante <b>CLOSE_COMMENT_MODE</b> correspond au mot cl dsignant la fermeture d'un bloc de commentaire.
   */
  private static final String CLOSE_COMMENT_MODE = "*/";
  
  //-----------------------
  // VARIABLES STATIQUES //
  //-----------------------
  
  /**
   * La variable static <b>comment_mode</b> dtermine si la lecture est en mode commentaire. 
   * En mode commentaire, toute les lignes lues ne sont pas traites tant que la variable <i>comment_mode</i> n'est pas <b>false</b>.
   */
  private static boolean comment_mode = false;
  
  //-------------
  // VARIABLES //
  //-------------
  
  /**
   * La variable <b>file_name</b> reprsente le nom du fichier qui sera utilis si l'objet de type SRW est crit dans un fichier texte.
   */
  private String file_name;
  
  /**
   * La variable <b>comment_list</b> reprsente une liste de commentaire qui sera inscrit avec l'criture de l'objet.
   */
  private final List<String> comment_list;
  
  //----------------
  // CONSTRUCTEUR //
  //----------------
  
  /**
   * ...
   */
  public SAbstractRW()
  {
    this(DEFAULT_FILE_NAME);
  }
  
  /**
   * ...
   * 
   * @param file_name
   */
  public SAbstractRW(String file_name)
  {
    this.file_name = file_name;
    comment_list = new ArrayList<String>();
  }
  
  //------------
  // MTHODES //
  //------------
  
  /**
   * Mthode pour obtenir le nom du fichier en lecture.
   * 
   * @return Le nom du fichier.
   */
  public String getFileName()
  {
    return file_name;
  }
  
  /**
   * Mthode pour dfinir le nom du fichier o sera lu ou crit le contenu de cet objet.
   * 
   * @param file_name Le nom du fichier.
   */
  public void setFileName(String file_name)
  {
    this.file_name = file_name;
  }
  
  @Override
  public String[] getRWParameterName()
  {
    return KEYWORD_PARAMETER;
  }
  
  @Override
  public final void read() throws SReadingException
  {
    read(new SBrowser());
  }
  
  /**
   * Mthode pour raliser la lecture dans un fichier de l'objet SRW  partir d'un rpertoire dfinie par le <i>browser</i>  l'aide de sa dfinition de nom de fichier.
   * 
   * @param browser Le <i>browser</i>
   * @throws SReadingException Si une erreur de lecture est survenue.
   */
  public final void read(SBrowser browser) throws SReadingException
  {
    if(file_name == null)
      throw new SReadingException("Erreur SAbstractRW  001 : L'objet de type '" + this.getClass().getTypeName() + "' ne peut pas tre lu dans un fichier txt, car le nom du fichier n'a pas t spcifi.");
    
    try {
      
      File file = browser.findFile(file_name);
      
      FileReader fr = new FileReader(file);
      SBufferedReader sbr = new SBufferedReader(fr);
      
      try{
        read(sbr);
      }catch(SInitializationException e){
        sbr.close();
        fr.close();
        throw new SReadingException("Erreur SAbstractRW 002 : Une erreur  l'initialisation est survenue." + SStringUtil.END_LINE_CARACTER + "\t" + e.getMessage(), e);
      }
      
      sbr.close();
      fr.close();
      
    }catch(FileNotFoundException e) {
      throw new SReadingException("Erreur SAbstractRW 003 : Le fichier '" + file_name + "' n'est pas trouv.", e);
    }catch(SManyFilesFoundException e) {
      throw new SReadingException("Erreur SAbstractRW 004 : Le fichier '" + file_name + "' a t trouv plus d'une fois dans les diffrents sous-rpertoires. Veuillez en garder qu'une seule version.", e);
    }catch(IOException e) {
      throw new SReadingException("Erreur SAbstractRW 005 : Le fichier '" + file_name + "' a rencontr une exception de type I/O.", e);
    }
    
  }
   
  /**
   * Mthode qui permet la lecture d'un fichier de format txt pour faire l'initialisation des paramtres de l'objet SAbstractReadable.
   * 
   * @param sbr Le buffer cherchant l'information dans le fichier txt.
   * @throws IOException Si une erreur de type I/O est survenue lors de la lecture.
   * @throws SInitializationException Si une erreur est survenue lors de l'initialisation de l'objet.
   */
  protected final void read(SBufferedReader sbr) throws SInitializationException 
  {
    int code;     //code du mot cl  analyser
    String line;  //ligne lu qui sera analyse dont le premier mot correspond au mot cl qui sera analys
    
    do
    {
      //initialisation de la lecture d'une ligne
      code = SKeyWordDecoder.CODE_NULL;   
      
      // Lecture de la ligne dans le fichier.
      try {
        line = sbr.readLine();
      }catch(IOException e) {
        throw new SInitializationException("Erreur SAbstractRW 006 : Une erreur est survenue lors de la lecture d'une ligne dans le fichier.", e);
      }
      
      //Vrification si le fichier est  sa dernire ligne sans avoir termin la lecture de l'objet courant.
      if(line == null)
      {
        // forcer la prsence du keyWord de fin de lecture de l'objet (oubli dans le fichier)
        code = SKeyWordDecoder.CODE_END;  
        
        // Si la lecture est en mode 'commentaire', il faudra retirer cette option
        // pour ne pas affecter la lecture d'un prochain fichier puisque ce paramtre est statique
        comment_mode = false;
      }
      else
      {
        StringTokenizer tokens = new StringTokenizer(line);   // Obtenir les tokens de la ligne  analyser
        
        //S'il y a des tokens  analyser
        if(tokens.countTokens() > 0)
        {
          String key_word = tokens.nextToken();       //obtenir le 1ier mot cl de la ligne
        
          //Vrifier si le mot dbut par un caratre signalant une ligne de commentaire
          if(!isCommentExpression(key_word))
          {
            code = SKeyWordDecoder.getKeyWordCode(key_word);  //gnr le code de rfrence
                  
            //Vrifier si la lecture de l'objet n'est pas complt
            if(code != SKeyWordDecoder.CODE_END)    
              //Vrifier si la lecture est un code d'erreur
              if(code == SKeyWordDecoder.CODE_ERROR)  
              {  
                // Afficher un message signalant que le "keyword" n'est pas reconnu par le systme (mot qui n'existe pas)
                SLog.logWriteLine("Ligne " + (sbr.atLine()-1) + " --- " + "Le mot cl '" + key_word + "' n'est pas reconnu par le systme.");
                
                // Afficher un message signalant une suggestion de mot cl valide dans la lecture courante de l'objet
                SLog.logWriteLine("Lecture en cours : " + getRWName());
                SLog.logWriteLine("Mots cls admissibles en paramtres : " + SStringUtil.join(getRWParameterName(), ", "));
              }
              else
              {
                //Retirer la prsence du mot cl  la ligne restante
                String remaining_line = SStringUtil.removeAllFirstSpacerCaracter(line);     //Retirer les 1ier caractres d'espacement de la ligne (s'il y en a) pour obtenir la ligne restante
                remaining_line = remaining_line.substring(key_word.length());               //Retirer le mot cl de la ligne restante  
                remaining_line = SStringUtil.removeAllFirstSpacerCaracter(remaining_line);  //Retirer les caractres d'espacement suivant le mot cl
                
                //Faire le traitement du mot cl :
                try{
                  
                  // Afficher un message d'erreur si le code ne fait pas parti de la dfinition de l'objet (mot dans le systme, mais pas pour cet objet).
                  if(!read(sbr,code,remaining_line))  
                  {
                    SLog.logWriteLine("Ligne " + (sbr.atLine()-1) + " --- " + "Le mot cl '" + key_word + "' ne fait pas parti de la dfinition de l'objet '" + getRWName() + "'."); 
                    SLog.logWriteLine("Mots cls admissibles en paramtres : " + SStringUtil.join(getRWParameterName(), ", "));
                    SLog.logWriteLine();
                  }
                
                }catch(SReadingException sre){
                  // Afficher un message d'erreur dsignant une erreur dans la lecture.
                  
                  // Si lecture dans un objet.
                  if(this.getFileName() == null)
                    SLog.logWriteLine("Ligne " + (sbr.atLine()-1) + " --- " + "Une erreur de lecture est survenue dans un objet '" + this.getRWName() + "'.");
                  else
                    SLog.logWriteLine("Ligne " + (sbr.atLine()-1) + " --- " + "Une erreur de lecture est survenue dans le fichier '" + this.getFileName() + "'.");
                  
                  SLog.logException(sre);
                }
              }
          }//fin if !iscommentWord
        }//fin if tokens > 0
      }//fin else line != null
    }while(code != SKeyWordDecoder.CODE_END);
    
    initialization(); 
  } 

  /**
   * Mthode pour initialiser l'objet aprs avoir complt la lecture. 
   * Cette mthode sera vide s'il n'y a pas de traitement de donne  effectuer sur les paramtres de lecture pour rendre l'objet oprationnel.
   * 
   * @throws SInitializationException Si une erreur est survenue lors de l'initialisation.
   */
  protected void initialization() throws SInitializationException
  {
    
  }
  
  /**
   * Mthode pour analyser le code du le mot cl lu dans le fichier txt et excuter la bonne action 
   * sur les paramtres de l'objet en considrant l'information contenue dans le restant 
   * d'une ligne lue ou en continuant la lecture dans le buffer.
   * @param sbr Le buffer de lecture.
   * @param code Le code du mot cl a excuter.
   * @param remaining_line Le restant de la ligne suivant le dernier mot cl lu.
   * @throws SReadingException Si une erreur de lecture a t dtecte dans le fichier.
   * @return <b>true</b> si le code a t trait par l'objet et <b>false</b> sinon.
   */
  protected boolean read(SBufferedReader sbr, int code, String remaining_line) throws SReadingException
  {
    switch(code)
    {
      case SKeyWordDecoder.CODE_FILE_NAME : file_name = this.readStringNotEmpty(remaining_line, SKeyWordDecoder.KW_FILE_NAME); return true;
                          
      case SKeyWordDecoder.CODE_COMMENT : comment_list.add(this.readStringNotEmpty(remaining_line, SKeyWordDecoder.KW_COMMENT)); return true;
          
      default : return false;
    }
  }
    
  @Override
  public final void write() throws IOException
  {
    write(this.file_name);
  }
    
  @Override
  public final void write(String file_name) throws IOException
  {
    write(file_name, new SBrowser());
  }
  
  /**
   * ...
   * 
   * @param browser
   * @throws IOException
   */
  public final void write(SBrowser browser) throws IOException
  {
    write(this.file_name, browser);
  }
  
  /**
   * Mthode pour crire un objet dans un fichier texte  partir d'un nom de fichier (pas ncessairement le mme que celui de la classe).
   * 
   * @param browser Le browser positionn  l'emplacement o le fichier sera crit.
   * @throws IOException S'il y a eu une erreur de type I/O.
   */
  public final void write(String file_name, SBrowser browser) throws IOException
  {
    // Vrification que l'objet de type SRW possde un nom de fichier pour s'crire dedans.
    if(file_name == null)
      throw new IOException("Erreur SAbstractRW  001 : L'objet de type '" + this.getClass().getTypeName() + "' ne peut pas tre crit dans un fichier txt, car le nom du fichier n'a pas t spcifi.");
    
    File file = browser.createFile(file_name);
    
    FileWriter fw = new FileWriter(file);
    BufferedWriter bw = new BufferedWriter(fw);
    
    // crire l'information contenu dans l'objet.
    this.writeInformation(bw);
    
    bw.close(); //  fermer celui-ci en premier, sinon, ERROR !!!
    fw.close(); 
  }
  
  @Override
  public void writeInformation(BufferedWriter bw) throws IOException
  {
    // criture du nom du fichier.
    if(this.file_name != null)
    {
      bw.write(SKeyWordDecoder.KW_FILE_NAME);
      bw.write("\t");
      bw.write(this.file_name);
      bw.write(SStringUtil.END_LINE_CARACTER);
      bw.write(SStringUtil.END_LINE_CARACTER);
    }
    
    // criture de la liste de commentaire.
    if(!this.comment_list.isEmpty())
    {
      for(String s : comment_list)
      {
        bw.write(SKeyWordDecoder.KW_COMMENT);
        bw.write("\t");
        bw.write(s);
        bw.write(SStringUtil.END_LINE_CARACTER);
        bw.write(SStringUtil.END_LINE_CARACTER);
      }
    }
        
  }
    
  /**
   * Mthode pour dterminer si le mot lue dsigne une ligne de commentaire.
   * Un mot de commentaire est <b>vide</b> ou dbute par un <b>caractre particulier</b> (#/*-!@)
   * ou bien la lecture est en mode commentaire (<i>comment_mode</i> = <b>true</b>).
   * @param word - Le mot  analyser.
   * @return <b>true</b> si la ligne est considre comme un commentaire et <b>false</b> sinon.
   */
  protected boolean isCommentExpression(String word)
  {
    //Vrification du mot vide (pour ne pas lire un premier caractre inexistant)
    if(word.equals(""))
      return true;
    
    //Recherche du mot cl "/*" ou "*/" pour changer l'tat de lecture (mode commentaire on/off)
    if(word.length() == 2)
      if(word.equals(OPEN_COMMENT_MODE))    // le mot /*
      {      
        //SLog.logWriteLine("Message SAbstractRW - Ouverture d'un bloc de commentaire.");
        comment_mode = true;
        return true;
      }
      else
        if(word.equals(CLOSE_COMMENT_MODE)) // le mot */
        {
          //SLog.logWriteLine("Message SAbstractRW - Fermeture d'un bloc de commentaire.");
          comment_mode = false;
          return true;
        }
    
    // Analyse du premier caractre de la ligne pour vrifier si seulement cette ligne est retire de l'analyse
    // seulement si la lecture n'est pas en mode commentaire. Dans ce cas, la ligne sera considre comme un commentaire.
    if(comment_mode)
      return true;
    else
    {
      char first_letter = word.charAt(0);
        
      for(int i = 0; i < COMMENT_LETTER.length(); i++)
        if(first_letter == COMMENT_LETTER.charAt(i))
          return true;
    }
        
    return false;
  }
  
  /**
   * Mthode pour faire l'analyse d'un string afin d'y retourner une valeur de type int.
   * @param remaining_line L'expression en String de la valeur.
   * @param value_name Le nom en mots correspondant  la dfinition de la valeur. Cette information sera utilise s'il y a une exception de lance.
   * @return La valeur numrique.
   * @throws SReadingException S'il y a une erreur de lecture.
   */
  protected final int readInt(String remaining_line, String value_name)throws SReadingException
  {
    StringTokenizer tokens = new StringTokenizer(remaining_line, SStringUtil.REMOVE_CARACTER_TOKENIZER);  
    
    if(tokens.countTokens() == 0)
      throw new SReadingException("Erreur SAbstractRW 015 : Il n'y a pas de valeur numrique affecte au '" + value_name + "'.");
    
    String s = tokens.nextToken();
      
    try{

      return Integer.valueOf(s);

    }catch(NumberFormatException e){ 
      throw new SReadingException("Erreur SAbstractRW 016 : L'expression '" + s + "' n'est pas un nombre de type int pouvant tre affect  la dfinition d'un '" + value_name + "'.", e);
    } 
  }
  
  /**
   * Mthode pour faire l'analyse d'un string afin d'y retourner une valeur de type int suprieure ou gale  0.
   * @param remaining_line - L'expression en String de la valeur.
   * @param value_name - Le nom en mots correspondant  la dfinition de la valeur. Cette valeur sera utilise s'il y a une exception de lance.
   * @return La valeur numrique.
   * @throws SReadingException S'il y a une erreur de lecture.
   */
  protected final int readIntEqualOrGreaterThanZero(String remaining_line, String value_name)throws SReadingException
  {
    StringTokenizer tokens = new StringTokenizer(remaining_line, SStringUtil.REMOVE_CARACTER_TOKENIZER);  
    
    if(tokens.countTokens() == 0)
      throw new SReadingException("Erreur SAbstractReadable 001 : Il n'y a pas de valeur numrique affecte au " + value_name +".");
    
    String s = tokens.nextToken();
        
    try{
      int value = Integer.valueOf(s);
      
      if(value < 0)
        throw new SReadingException("Erreur SAbstractReadable 002 : Le " + value_name + " '" + value + "' doit tre gal ou suprieur  0.");
      else
        return value;
    }catch(NumberFormatException e){ 
      throw new SReadingException("Erreur SAbstractReadable 003 : L'expression '" + s + "' n'est pas un nombre de type int pouvant tre affect  la dfinition d'un " + value_name + ".");
    } 
  }
  
  /**
   * Mthode pour faire l'analyse d'un string afin d'y retourner une valeur de type int suprieure  0.
   * 
   * @param remaining_line - L'expression en String de la valeur.
   * @param value_name - Le nom en mots correspondant  la dfinition de la valeur. Cette information sera utilise s'il y a une exception de lance.
   * @return La valeur numrique.
   * @throws SReadingException S'il y a une erreur de lecture.
   */
  protected final int readIntGreaterThanZero(String remaining_line, String value_name)throws SReadingException
  {
    StringTokenizer tokens = new StringTokenizer(remaining_line, SStringUtil.REMOVE_CARACTER_TOKENIZER);
    
    if(tokens.countTokens() == 0)
      throw new SReadingException("Erreur SAbstractReadable 004 : Il n'y a pas de valeur numrique affecte au '" + value_name +"'.");
    
    String s = tokens.nextToken();
        
    try{
      int value = Integer.valueOf(s);
      
      if(value < 1)
        throw new SReadingException("Erreur SAbstractReadable 005 : Le '" + value_name + "' de valeur '" + value + "' doit tre suprieur  0.");
      else
        return value;
      
    }catch(NumberFormatException e){ 
      throw new SReadingException("Erreur SAbstractReadable 006 : L'expression '" + s + "' n'est pas un nombre de type int pouvant tre affect  la dfinition d'un '" + value_name + "'.");
    } 
  }
  
  /**
   * Mthode pour faire l'analyse d'un string afin d'y retourner une valeur de type int suprieure ou gale  une valeur de rfrence.
   * 
   * @param remaining_line L'expression en String de la valeur.
   * @param value La valeur de rfrence.
   * @param value_name Le nom en mots correspondant  la dfinition de la valeur. Cette information sera utilise s'il y a une exception de lance.
   * @return La valeur numrique.
   * @throws SReadingException S'il y a une erreur de lecture.
   */
  protected final int readIntEqualOrGreaterThanValue(String remaining_line, int value, String value_name) throws SReadingException
  {
    int cst = readInt(remaining_line, value_name);
  
    if(cst < value)
      throw new SReadingException("Erreur SAbstractRW 018 : La valeur '" + cst +"' est infrieure  la valeur minimale acceptable tant '" + value +" pouvant tre affect  la dfinition d'un '" + value_name + "'.");
    
    return cst;
  }
  
  /**
   * Mthode pour faire l'analyse d'un string afin d'y retourner une valeur de type integer suprieure ou gale  une valeur de rfrence et infrieure ou gale  une valeur de rfrence.
   *
   * @param remaining_line L'expression en String de la valeur.
   * @param min_value La valeur de rfrence minimale.
   * @param max_value La valeur de rfrence maximale.
   * @param value_name Le nom en mots correspondant  la dfinition de la valeur. Cette information sera utilise s'il y a une exception de lance.
   * @return La valeur numrique.
   * @throws SReadingException S'il y a une erreur de lecture.
   * @throws SRuntimeException Si la valeur de rfrence minimale et maximal ne sont pas cohrentes.
   */
  protected final int readIntEqualOrGreaterThanValueAndEqualOrSmallerThanValue(String remaining_line, int min_value, int max_value, String value_name) throws SReadingException, SRuntimeException
  {
    if(min_value > max_value)
      throw new SRuntimeException("Erreur SAbstractReadable 019 : L'usage de la mthode est erron. La valeur min_value = " + min_value + " est suprieure  max_value = " + max_value + ".");
    
    int value = readIntEqualOrGreaterThanValue(remaining_line, min_value, value_name);
    
    if(value > max_value)
      throw new SReadingException("Erreur SAbstractReadable 020 : La valeur '" + value +"' est suprieure  la valeur maximale acceptable tant '" + max_value +" pouvant tre affect  la dfinition d'un '" + value_name + "'.");
    
    return value;
  }
  
  /**
   * Mthode pour faire l'analyse d'un string afin d'y retourner une valeur de type int gale ou suprieure  0. 
   * Une expression en String peut tre galement reconnue pour dfinir la valeur de type int dsire.
   * @param remaining_line - L'expression en String de la valeur.
   * @param value_name - Le nom en mots correspondant  la dfinition de la valeur. Cette information sera utilise s'il y a une exception de lance.
   * @param valid_expression - Le tableau des String pouvant tre reconnue. Leur position dans le tableau dfinisse galement leur valeur attitre.
   * @return La valeur numrique.
   * @throws SReadingException S'il y a une erreur de lecture.
   */
  protected final int readIntOrExpression(String remaining_line, String value_name, String[] valid_expression)throws SReadingException
  {
    StringTokenizer tokens = new StringTokenizer(remaining_line, SStringUtil.REMOVE_CARACTER_TOKENIZER);  
    
    if(tokens.countTokens() == 0)
      throw new SReadingException("Erreur SAbstractReadable 007 : Il n'y a pas de valeur numrique affecte au '" + value_name + "'.");
   
    String s = tokens.nextToken();
    
    try{ 
      
      //Analyse par une valeur numrique :
      int value = Integer.valueOf(s);
      
      if(value < 0)
        throw new SReadingException("Erreur SAbstractReadable 008 : Le " + value_name + " '" + value + "' doit tre gal ou suprieur  0.");
      else
        if(value >= valid_expression.length) //ne pas dpasser le nombre d'expression valide
          throw new SReadingException("Erreur SAbstractReadable 009 : Le " + value_name + " '" + value + "' doit tre infrieur  '" + valid_expression.length + "'. Les expressions suivantes sont galements valides : " + SStringUtil.join(valid_expression, ", "));
        else
          return value;
      
    }catch(NumberFormatException e){ 
      
      //Analyse par mot cl : Recherche du mot cl dans la liste des expressions valides dont l'index correspond  la valeur numrique attitre
      for(int i=0; i<valid_expression.length; i++)
        if(s.equalsIgnoreCase(valid_expression[i]))
          return i;
      
      throw new SReadingException("Erreur SAbstractReadable 010 : L'expression '" + s + "' n'est pas reconnue pour dfinir la valeur '" + value_name + "'. Les expressions valides sont les suivantes : " + SStringUtil.join(valid_expression, ", "));
    }
  }
        
  /**
   * Mthode pour faire l'analyse d'un string afin d'y retourner un string <b>non vide</b>.
   * @param remaining_line - L'expression en String de la valeur.
   * @param value_name - Le nom en mots correspondant  la dfinition de la valeur. Cette information sera utilise s'il y a une exception de lance.
   * @return Le string non vide.
   * @throws SReadingException S'il y a une erreur de lecture.
   */
  protected final String readStringNotEmpty(String remaining_line, String value_name)throws SReadingException
  {
    if(remaining_line.length() == 0)
      throw new SReadingException("Erreur SAbstractReadable 011 : Le string  analyser est '" + remaining_line + "' et n'est pas une expression adquate  affecteur  la valeur '" + value_name + "'.");
    else
      return remaining_line;
  }
  
  /**
   * Mthode pour faire l'analyse d'un string dsignant <b>vrai</b> ou <b>faux</b>. 
   * Un nombre entier peut galement tre lu o <b>0</b> dsigne <b>faux</b> et les <b>autres entiers positifs</b> dsigne <b>vrai</b>.
   * @param remaining_line - L'expression en String de la valeur.
   * @param value_name - Le nom en mots correspondant  la dfinition de la valeur. Cette information sera utilise s'il y a une exception de lance.
   * @return <b>true</b> si la lecture est interprte comme <b>vrai</b> et <b>false</b> si la lecture est interprte comme tant <b>faux</b>.
   * @throws SReadingException S'il y a une erreur de lecture.
   */
  protected final boolean readTrueFalseExpressionOrInt(String remaining_line, String value_name) throws SReadingException
  {
    StringTokenizer tokens = new StringTokenizer(remaining_line, SStringUtil.REMOVE_CARACTER_TOKENIZER);  
    
    if(tokens.countTokens() == 0)
      throw new SReadingException("Erreur SAbstractReadable 012 : Il n'y a pas de valeur numrique affecte au '" + value_name + "'.");
    
    String s = tokens.nextToken();
    
    try{ 
      
      //Analyse par une valeur numrique :
      int value = Integer.valueOf(s);
      
      if(value < 0)
        throw new SReadingException("Erreur SAbstractReadable 013 : Le " + value_name + " '" + value + "' doit tre gal ou suprieur  0.");
      else
        if(value == 0)
          return false;
        else
          return true;
      
    }catch(NumberFormatException e){ 
      
      //Analyse par mot cl : true/vrai/yes/oui et false/faux/no/non
      s = s.toLowerCase(Locale.ENGLISH);
      
      for(int i = 0; i < TRUE_EXPRESSION.length; i++)
        if(s.equals(TRUE_EXPRESSION[i]))
          return true;
      
      for(int i = 0; i < FALSE_EXPRESSION.length; i++)
        if(s.equals(FALSE_EXPRESSION[i]))
          return false;
      
      throw new SReadingException("Erreur SAbstractReadable 014 : L'expression '" + s + "' n'est pas reconnue pour dfinir la valeur '" + value_name + "'. Les expressions valides sont les suivantes : " + SStringUtil.join(TRUE_EXPRESSION, ", ") + " et " + SStringUtil.join(FALSE_EXPRESSION, ", "));
    }
  }
  
  /**
   * Mthode pour faire l'analyse d'un string afin d'y retourner une valeur de type double.
   * @param remaining_line - L'expression en String de la valeur.
   * @param value_name - Le nom en mots correspondant  la dfinition de la valeur. Cette information sera utilise s'il y a une exception de lance.
   * @return La valeur numrique.
   * @throws SReadingException S'il y a une erreur de lecture.
   */
  protected final double readDouble(String remaining_line, String value_name)throws SReadingException
  {
    StringTokenizer tokens = new StringTokenizer(remaining_line, SStringUtil.REMOVE_CARACTER_TOKENIZER);  
    
    if(tokens.countTokens() == 0)
      throw new SReadingException("Erreur SAbstractReadable 015 : Il n'y a pas de valeur numrique affecte au '" + value_name + "'.");
    
    String s = tokens.nextToken();
      
    try{

      return Double.valueOf(s);

    }catch(NumberFormatException e){ 
      throw new SReadingException("Erreur SAbstractReadable 016 : L'expression '" + s + "' n'est pas un nombre de type double pouvant tre affect  la dfinition d'un '" + value_name + "'.", e);
    } 
  }
  
  /**
   * Mthode pour faire l'analyse d'un string afin d'y retourner une valeur de type double suprieure  0 (et non nul).
   * @param remaining_line - L'expression en String de la valeur.
   * @param value_name - Le nom en mots correspondant  la dfinition de la valeur. Cette information sera utilise s'il y a une exception de lance.
   * @return La valeur numrique.
   * @throws SReadingException S'il y a une erreur de lecture.
   */
  protected final double readDoubleGreaterThanZero(String remaining_line, String value_name)throws SReadingException
  {
    double value = readDouble(remaining_line, value_name);
    
    if(value < SMath.EPSILON)
      throw new SReadingException("Erreur SAbstractReadable 017 : Le '" + value_name + "' de valeur '" + value + "' doit tre suprieur  '" + SMath.EPSILON + "' pour tre valide.");
     
    return value;
  }
  
  /**
   * Mthode pour faire l'analyse d'un string afin d'y retourner une valeur de type double suprieur ou gale  une valeur de rfrence.
   * @param remaining_line L'expression en String de la valeur.
   * @param value La valeur de rfrence.
   * @param value_name Le nom en mots correspondant  la dfinition de la valeur. Cette information sera utilise s'il y a une exception de lance.
   * @return La valeur numrique.
   * @throws SReadingException S'il y a une erreur de lecture.
   */
  protected final double readDoubleEqualOrGreaterThanValue(String remaining_line, double value, String value_name) throws SReadingException
  {
    double cst = readDouble(remaining_line, value_name);
  
    if(cst < value)
      throw new SReadingException("Erreur SAbstractReadable 018 : La valeur '" + cst +"' est infrieure  la valeur minimale acceptable tant '" + value +" pouvant tre affect  la dfinition d'un '" + value_name + "'.");
    
    return cst;
  }
  
  /**
   * Mthode pour faire l'analyse d'un string afin d'y retourner une valeur de type double suprieure ou gale  une valeur de rfrence et infrieure ou gale  une valeur de rfrence.
   *
   * @param remaining_line L'expression en String de la valeur.
   * @param min_value La valeur de rfrence minimale.
   * @param max_value La valeur de rfrence maximale.
   * @param value_name Le nom en mots correspondant  la dfinition de la valeur. Cette information sera utilise s'il y a une exception de lance.
   * @return La valeur numrique.
   * @throws SReadingException S'il y a une erreur de lecture.
   * @throws SRuntimeException Si la valeur de rfrence minimale et maximal ne sont pas cohrentes.
   */
  protected final double readDoubleEqualOrGreaterThanValueAndEqualOrSmallerThanValue(String remaining_line, double min_value, double max_value, String value_name) throws SReadingException, SRuntimeException
  {
    if(min_value > max_value)
      throw new SRuntimeException("Erreur SAbstractReadable 019 : L'usage de la mthode est erron. La valeur min_value = " + min_value + " est suprieure  max_value = " + max_value + ".");
    
    double value = readDoubleEqualOrGreaterThanValue(remaining_line, min_value, value_name);
    
    if(value > max_value)
      throw new SReadingException("Erreur SAbstractReadable 020 : La valeur '" + value +"' est suprieure  la valeur maximale acceptable tant '" + max_value +" pouvant tre affect  la dfinition d'un '" + value_name + "'.");
    
    return value;
  }
  
  /**
   * Mthode pour faire l'analyse d'un string afin d'y retourner un tableau de type double.
   * @param remaining_line L'expression en String de la valeur.
   * @param value_name Le nom en mots correspondant  la dfinition de la valeur. Cette information sera utilise s'il y a une exception de lance.
   * @return Le tableau de type double.
   * @throws SReadingException S'il y a une erreur de lecture.
   */
  protected final double[] readDoubleArray(String remaining_line, String value_name)throws SReadingException
  {
    StringTokenizer tokens = new StringTokenizer(remaining_line, SStringUtil.REMOVE_CARACTER_TOKENIZER);  
    
    if(tokens.countTokens() == 0)
      throw new SReadingException("Erreur SAbstractReadable 021 : Il n'y a pas de valeur numrique affecte au '" + value_name + "'.");
    
    double[] array = new double[tokens.countTokens()];
    
    String s;
    int i = 0;
    
    while(tokens.hasMoreTokens())      
    {
      s = tokens.nextToken();
      
      try{
        array[i] = Double.valueOf(s);
        i++;
      }catch(NumberFormatException e){ 
        throw new SReadingException("Erreur SAbstractReadable 022 : L'expression '" + s + "' n'est pas un nombre de type double pouvant tre affect  la dfinition d'un '" + value_name + "'.", e);
      } 
    }
    
    return array; 
  }
  
  /**
   * Mthode pour faire l'analyse d'un string afin d'y retourner un tableau de type int.
   * 
   * @param remaining_line L'expression en String de la valeur.
   * @param value_name Le nom en mots correspondant  la dfinition de la valeur. Cette information sera utilise s'il y a une exception de lance.
   * @return Le tableau de type int.
   * @throws SReadingException S'il y a une erreur de lecture.
   */
  protected final int[] readIntArray(String remaining_line, String value_name)throws SReadingException
  {
    StringTokenizer tokens = new StringTokenizer(remaining_line, SStringUtil.REMOVE_CARACTER_TOKENIZER);  
    
    if(tokens.countTokens() == 0)
      throw new SReadingException("Erreur SAbstractRW 022 : Il n'y a pas de valeur numrique affecte au '" + value_name + "'.");
    
    int[] array = new int[tokens.countTokens()];
    
    String s;
    int i = 0;
    
    while(tokens.hasMoreTokens())      
    {
      s = tokens.nextToken();
      
      try{
        array[i] = Integer.valueOf(s);
        i++;
      }catch(NumberFormatException e){ 
        throw new SReadingException("Erreur SAbstractRW 024 : L'expression '" + s + "' n'est pas un nombre de type int pouvant tre affect  la dfinition d'un '" + value_name + "'.", e);
      } 
    }
    
    return array; 
  }
  
  /**
   * Mthode pour faire l'analyse d'un string afin d'y retourner un objet de type SVector3d.
   * 
   * @param remaining_line L'expression en String de la valeur.
   * @param value_name Le nom en mots correspondant  la dfinition de la valeur. Cette information sera utilise s'il y a une exception de lance.
   * @return Le vecteur
   * @throws SReadingException S'il y a une erreur de lecture.
   */
  protected final SVector3d readSVector3d(String remaining_line, String value_name) throws SReadingException
  {
    try {
      
      return new SVector3d(remaining_line);
      
    }catch(SConstructorException e) {
      throw new SReadingException("Erreur SAbstractRW 025 : L'expression '" + remaining_line + "' n'est pas valide pour dfinir un vecteur.", e);
    }
  }

  /**
   * Mthode pour faire l'analyse d'un string afin d'y retourner un objet de type SColor.
   * 
   * @param remaining_line L'expression en String de la valeur.
   * @param value_name Le nom en mots correspondant  la dfinition de la valeur. Cette information sera utilise s'il y a une exception de lance.
   * @return La couleur.
   * @throws SReadingException S'il y a une erreur de lecture.
   */
  protected final SColor readSColor(String remaining_line, String value_name) throws SReadingException
  {
    try {
      
      return new SColor(remaining_line);
      
    }catch(SConstructorException e) {
      throw new SReadingException("Erreur SAbstractRW 025 : L'expression '" + remaining_line + "' n'est pas valide pour dfinir une couleur.", e);
    }
  }
  
  /**
   * Mthode pour faire l'analyse d'un string afin d'y retourner un objet de type SVectorUV.
   * 
   * @param remaining_line L'expression en String de la valeur.
   * @param value_name Le nom en mots correspondant  la dfinition de la valeur. Cette information sera utilise s'il y a une exception de lance.
   * @return Le vecteur
   * @throws SReadingException S'il y a une erreur de lecture.
   */
  protected final SVectorUV readSVectorUV(String remaining_line, String value_name) throws SReadingException
  {
    try {
      
      return new SVectorUV(remaining_line);
      
    }catch(SConstructorException e) {
      throw new SReadingException("Erreur SAbstractRW 025 : L'expression '" + remaining_line + "' n'est pas valide pour dfinir un vecteur", e);
    }
  }
  
}
