/**
 * 
 */
package sim.readwrite.object;

import java.io.BufferedWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import sim.physics.SGravitationnalParticleSystem;
import sim.physics.SParticleSystem;
import sim.readwrite.SAbstractRW;
import sim.util.SBufferedReader;
import sim.util.SInitializationException;
import sim.util.SKeyWordDecoder;
import sim.util.SReadingException;
import sim.util.SStringUtil;

/**
 * La classe <b>SParticlesSystemRW</b> reprsente un systme de particule pouvant tre lue dans un fichier texte.
 * 
 * @author Simon Vzina
 * @since 2023-04-23
 * @version 2023-09-13 (Version : Les orbites (2023-10-30))
 */
public class SParticleSystemRW extends SAbstractRW {

  /**
   * Le paramtre <b>particles_reading_list<b> reprsente la liste des particules lues.
   */
  private final List<SParticleRW> particles_list;
  
  /**
   * La constante <b>KEYWORD_PARAMETER</b> correspond  un tableau contenant l'ensemble des mots cls 
   *  utiliser reconnus lors de la dfinition de l'objet par une lecture en fichier.
   */
  private static final String[] KEYWORD_PARAMETER = { 
      SKeyWordDecoder.KW_PARTICLE 
  };
  
  /**
   * Constructeur d'un systme de particule par dfaut.
   */
  public SParticleSystemRW()
  {
    particles_list = new ArrayList<SParticleRW>();
  }
      
  /**
   * Constructeur d'un systme de particule  partir fichier de lecture. 
   */
  public SParticleSystemRW(String file_name)
  {
    super(file_name);
    
    particles_list = new ArrayList<SParticleRW>();
  }
    
  /**
   * Constructeur d'un systme de particule  partir d'information lue dans un fichier de format txt.
   * 
   * @param sbr Le BufferedReader cherchant l'information dans le fichier txt.
   * @throws SReadingException Si une erreur est survenue  la construction.
   * @see SBufferedReader
   */
  public SParticleSystemRW(SBufferedReader sbr) throws SReadingException
  {
    this();   
    
    try{
      read(sbr);
    }catch(SInitializationException e){
      throw new SReadingException("Erreur SParticleSystemRW 001 : Une erreur d'initialisation est survenue.", e);
    }  
  }
    
  /**
   * Mthode pour convertir un SParticlesSystemRW en SParticlesSystem.
   * 
   * @return Le systme de particules.
   */
  public SParticleSystem toSGravitationnalParticleSystem()
  {
    return fillParticleSystem(new SGravitationnalParticleSystem());
  }
  
  /**
   * Mthode pour remplir un systme de particules.
   *  
   * @param sys Le systme.
   * @return Le systme de particule.
   */
  private SParticleSystem fillParticleSystem(SParticleSystem sys)
  {
    // Prendre la liste des particules et remplir le systme.
    for(SParticleRW p : this.particles_list)
      sys.add(p.toSParticle());
    
    return sys;
  }
  
  /**
   * Mthode pour obtenir la liste des particules.
   * 
   * @return La liste des particules.
   */
  public List<SParticleRW> getParticleRWList()
  {
    return this.particles_list;
  }
  
  @Override
  public void writeInformation(BufferedWriter bw) throws IOException
  {
    super.writeInformation(bw);
    
    // Faire l'criture de toutes les particules.
    for(SParticleRW p : particles_list)
      p.write(bw);
    
    bw.write(SStringUtil.END_LINE_CARACTER);
  }
  
  @Override
  protected boolean read(SBufferedReader sbr, int code, String remaining_line) throws SReadingException
  {
    switch(code)
    {
      case SKeyWordDecoder.CODE_PARTICLE : particles_list.add(new SParticleRW(sbr)); return true;
      
      default : return super.read(sbr, code, remaining_line);
    }
  }
  
  @Override
  protected void initialization() throws SInitializationException
  {
   
  }
  
  @Override
  public String getRWName() 
  {
    return SKeyWordDecoder.KW_PARTICLES_SYSTEM;
  }

  @Override
  public String[] getRWParameterName()
  {
    String[] other_parameters = super.getRWParameterName();
    
    return SStringUtil.merge(other_parameters, KEYWORD_PARAMETER);
  }
  
}
