package sim.math;

import org.junit.Assert;
import org.junit.Test;

import sim.exception.SNoImplementationException;
import sim.util.SLog;

/**
 * JUnit test permettant de valider les fonctionnalits de la classe <b>SMatrix3x3</b>.
 * 
 * @author Simon Vzina
 * @since 2020-05-05
 * @version 2022-04-13
 */
public class SMatrix3x3Test {

  /**
   * Test de la mthode <b>determinant</b> avec un matrice quelconque.
   * 
   * Rfrence : https://ncalculators.com/matrix/inverse-matrix.htm
   */
  @Test
  public void determinantTest1()
  {
    SMatrix3x3 M = new SMatrix3x3(
        
        10.0, 20.0, 10.0,
        4.0, 5.0, 6.0,
        2.0, 3.0, 5.0
        
        );
    
    Assert.assertEquals(-70.0, M.determinant(), 0.00001);
  }

  /**
   * Test de la mthode <b>inverse</b> avec un matrice quelconque.
   * 
   * Rfrence : https://ncalculators.com/matrix/inverse-matrix.htm
   */
  @Test
  public void inverseTest1()
  {
    SMatrix3x3 M = new SMatrix3x3(
        
        10.0, 20.0, 10.0,
        4.0, 5.0, 6.0,
        2.0, 3.0, 5.0
        
        );
   
    SMatrix3x3 expected = new SMatrix3x3(
        
        -7.0/70.0, 70.0/70.0, -70.0/70.0,
        8.0/70.0, -30.0/70.0, 20.0/70.0,
        -2.0/70.0, -10.0/70.0, 30.0/70.0
        
        );
    
    Assert.assertEquals(expected, M.inverse());
  }
    
  /**
   * Test de la mthode <b>skewSymmetric</b> dans le cas d'un calcul avec son propre vecteur se qui donnera zro.
   */
  @Test
  public void skewSymmetricTest1a()
  {
    try {
      SVector3d a = new SVector3d(2.0, -5.0, 7.0);
      SMatrix3x3 A = SMatrix3x3.skewSymmetric(a);
      
      Assert.assertEquals(SVector3d.ZERO, A.multiply(a));
      
    }catch(SNoImplementationException e){
      SLog.logWriteLine("SMatrix3x3Test ---> Test non effectu : public void skewSymmetricTest1a()");
    }
  }
  
  /**
   * Test de la mthode <b>skewSymmetric</b> pour tester la proprit : B*S(a) x a = 0 o B est une matrice quelconque.
   */
  @Test
  public void skewSymmetricTest1b()
  {
    try {
      SVector3d a = new SVector3d(2.0, -5.0, 7.0);
      SMatrix3x3 A = SMatrix3x3.skewSymmetric(a);
      
      SMatrix3x3 B = new SMatrix3x3(3.4, -2.3, 5.6,
                                   -9.2, 8.2, -2.3,
                                    7.7, -2.9, -5.9);
          
      Assert.assertEquals(SVector3d.ZERO, B.multiply(A).multiply(a));
    
    }catch(SNoImplementationException e){
      SLog.logWriteLine("SMatrix3x3Test ---> Test non effectu : public void skewSymmetricTest1b()");
    }
  }
  
  /**
   * Test de la mthode <b>skewSymmetric</b> dans la validation d'un calcul de produit vectoriel.
   */
  @Test
  public void skewSymmetricTest2()
  {
    try {
      SVector3d a = new SVector3d(2.0, -5.0, 7.0);
      SVector3d b = new SVector3d(3.0, 7.0, -4.0);
      
      Assert.assertEquals(a.cross(b), SMatrix3x3.skewSymmetric(a).multiply(b));
    
    }catch(SNoImplementationException e){
      SLog.logWriteLine("SMatrix3x3Test ---> Test non effectu : public void skewSymmetricTest2()");
    }
  }
    
  /**
   * Test de la mthode <b>skewSymmetric</b> dans la proprit (SB)^T = B^T S^T
   * 
   * Puisque S^T = -S, nous pouvons galement valider (SB)^T = - B^T S
   */
  @Test
  public void skewSymmetricTest3()
  {
    try {
      SVector3d a = new SVector3d(-3.0, 7.0, -4.0);
      
      SMatrix3x3 R = new SMatrix3x3(3.0, 2.0, -1.0,
                                    9.0, 8.0, 3.0,
                                    -5.0, -7.0, -2.0);
      
      // Le ct gauche de l'quation.
      SMatrix3x3 actual = SMatrix3x3.skewSymmetric(a).multiply(R).transpose();
      
      // Validation de (SB)^T = B^T S^T.
      SMatrix3x3 expected1 = R.transpose().multiply(SMatrix3x3.skewSymmetric(a).transpose());
      Assert.assertEquals(expected1, actual);
      
      // Validation de (SB)^T = -B^T S.
      SMatrix3x3 expected2 = R.transpose().multiply(SMatrix3x3.skewSymmetric(a).multiply(-1.0));
      Assert.assertEquals(expected2, SMatrix3x3.skewSymmetric(a).multiply(R).transpose());
    
    }catch(SNoImplementationException e){
      SLog.logWriteLine("SMatrix3x3Test ---> Test non effectu : public void skewSymmetricTest3()");
    }
  }
  
  /**
   * Test de la mthode <b>skewSymmetric</b> dans l'une des proprits du calcul de la drive
   * du moment d'inertie par rapport au temps.
   * 
   * Rfrence : https://physics.stackexchange.com/questions/530248/first-derivative-of-the-moment-of-inertia-in-time-as-a-physical-parameter
   */
  @Test
  public void skewSymmetricTest4()
  {
    try {
      
      SMatrix3x3 I_S = new SMatrix3x3(3.0, 2.0, -1.0,
                                      9.0, 8.0, 3.0,
                                      -5.0, -7.0, -2.0);
      
      SQuaternion q = new SQuaternion(1.2, new SVector3d(4.5, 3.0, 7.0).normalize());
      
      SMatrix3x3 R = q.rotationMatrix3x3();
      SMatrix3x3 R_inv = R.inverse();
      SVector3d w = new SVector3d(-3.0, 7.0, 6.0);
      
      SMatrix3x3 I_O = R.multiply(I_S).multiply(R_inv);
      
      SMatrix3x3 T1 = w.toSkewSymmetricMatrix().multiply(R).multiply(I_S).multiply(R_inv);
      SMatrix3x3 T2 = R.multiply(I_S).multiply(R_inv).multiply(w.toSkewSymmetricMatrix());
      
      SVector3d expected = T1.substract(T2).multiply(w);
      SVector3d actual = w.toSkewSymmetricMatrix().multiply(I_O).multiply(w);
      
      Assert.assertEquals(expected, actual);
    
    }catch(SNoImplementationException e){
      SLog.logWriteLine("SMatrix3x3Test ---> Test non effectu : public void skewSymmetricTest4()");
    }
  }
}
