/**
 * 
 */
package sim.physics;

import java.util.Objects;
import java.util.concurrent.atomic.AtomicLong;

import sim.exception.SConstructorException;
import sim.math.SVector3d;

/**
 * La classe <b>SParticle</b> reprsente une particule.
 *  
 * @author Simon Vezina
 * @since 2020-07-29
 * @version 2023-10-25
 */
public class SParticle {

  /**
   * La constante <b>DEFAULT_ELECTRIC_CHARGE</b> reprsente la charge lectrique d'une particule par dfaut tant gal  {@value}.
   */
  protected static final double DEFAULT_ELECTRIC_CHARGE = 0.0;
  
  /**
   * La constante <b>DEFAULT_SPEED</b> reprsente la vitesse par dfaut d'une particule tant gale  zro.
   */
  protected final static SVector3d DEFAULT_SPEED = SVector3d.ZERO;
  
  /**
   * La variable <b>atomic_counteur_ID</b> correspond  un compteur <i>THREAD PROOF</i>  numro d'identification des particules. 
   */
  private static final AtomicLong atomic_counter_ID = new AtomicLong();
  
  /**
   * Le paramtre <b>id</b> reprsente un numro d'identification unique pour une particule.
   */
  private final long id;
  
  /**
   * Le paramtre <b>mass</b> reprsente la masse de la particule.
   */
  private final double mass;
  
  /**
   * Le paramtre <b>electric_charge</b> reprsente la charge lectrique de la particule.
   */
  private final double electric_charge;
  
  /**
   * Le paramtre <b>component</b> reprsente les composants cintiques actuelle de la particule (position et vitesse initiale).
   */
  private SKineticComponent component;
  
  /**
   * Le paramtre <b>predicted_component</b> reprsente les composants cintiques prdits de la particule (position et vitesse prdites).
   * On peut galement exploiter les composant prdits comme tant les composants finaux associs  une cinmatique de la particule.
   */
  private SKineticComponent predicted_component;
  
  /**
   * Constructeur d'une particule  partir de la masse et sa position.
   * 
   * @param mass La masse.
   * @param position La position.
   * @throws SConstructorException Si la masse est ngative.
   */
  public SParticle(double mass, SVector3d position) throws SConstructorException
  {
    this(mass, DEFAULT_ELECTRIC_CHARGE, position, DEFAULT_SPEED);
  }
  
  /**
   * Constructeur d'une particule  partir de la masse, la position et la vitesse.
   * 
   * @param mass La masse.
   * @param position La position.
   * @param speed La vitesse.
   * @throws SConstructorException Si la masse est ngative.
   */
  public SParticle(double mass, SVector3d position, SVector3d speed) throws SConstructorException
  {
    this(mass, DEFAULT_ELECTRIC_CHARGE, position, speed);
  }
  
  /**
   * Constructueur d'une particule  partir de la masse, la charge et sa position.
   * 
   * @param mass La masse.
   * @param charge La charge lectrique.
   * @param position La position.
   * @throws SConstructorException Si la masse est ngative.
   */
  public SParticle(double mass, double charge, SVector3d position) throws SConstructorException
  {
    this(mass, charge, position, DEFAULT_SPEED);
  }
  
  /**
   * Constructeur d'une particule.
   * 
   * @param mass La masse.
   * @param charge La charge lectrique.
   * @param position La position.
   * @param speed La vitesse.
   * @throws SConstructorException Si la masse est ngative.
   */
  public SParticle(double mass, double charge, SVector3d position, SVector3d speed) throws SConstructorException
  {
    if(mass < 0.0)
      throw new SConstructorException("Erreur SParticle 001 : La masse ne peut pas tre ngative.");
    
    // Identification d'un numro d'identifiant unique.
    this.id = atomic_counter_ID.incrementAndGet();
        
    this.mass = mass;
    this.electric_charge = charge;
    
    //  la construction, les composants actuels et prdit seront identiques.
    this.component = new SKineticComponent(position, speed);
    this.predicted_component = this.component;
  }
  
  
  /**
   * Mthode pour obtenir le numro d'identification unique ID de la particule.
   * 
   * @return Le numro d'identification ID.
   */
  public long getID()
  {
    return this.id;
  }
  
  /**
   * Mthode pour obtenir la masse de la particule.
   * 
   * @return La masse.
   */
  public double getMass()
  {
    return this.mass;
  }
  
  /**
   * Mthode pour obtenir la charge lectrique d'une particule.
   * 
   * @return La charge lectrique.
   */
  public double getElectricCharge()
  {
    return this.electric_charge;
  }
  
  /**
   * Mthode pour obtenir la position de la particule.
   * 
   * @return La position.
   */
  public SVector3d getPosition()
  {
    return this.component.getPosition();
  }
  
  /**
   * Mthode pour obtenir la vitesse de la particule.
   * 
   * @return La vitesse.
   */
  public SVector3d getSpeed()
  {
    return this.component.getSpeed();
  }
  
  /**
   * Mthode pour obtenir les composants cintiques de la particule.
   * 
   * @return
   */
  public SKineticComponent getKineticComponent()
  {
    return this.component;
  }
  
  /**
   * Mthode pour dfinir la position de la particule.
   * 
   * @param position La position.
   */
  public void setPosition(SVector3d position)
  {
    this.setKineticComponent(new SKineticComponent(position, this.getSpeed()));
  }
  
  /**
   * Mthode pour dfinir la vitesse de la particule.
   * 
   * @param speed La vitesse.
   */
  public void setSpeed(SVector3d speed)
  {
    this.setKineticComponent(new SKineticComponent(this.getPosition(), speed));
  }
  
  /**
   * Mthode pour dfinir les composants cintiques de la particule.
   * 
   * @param c Les composants.
   */
  public void setKineticComponent(SKineticComponent c)
  {
    this.component = c;
  }
  
  /**
   * Mthode pour dfinir les composants cintiques prdits de la particule. 
   * 
   * @param c Les composants prdits.
   */
  public void setKineticPredictedComponent(SKineticComponent c)
  {
    this.predicted_component = c;
  }
  
  /**
   * <p>
   * Mthode pour permutter les composants cintiques de actuel  prdit ou de prdit  actuel.
   * </p>
   * <p>
   * Cette mthode peut tre utiliser pour dplacer une particule si la mthode <i>setKineticPredictedComponent(...)</i>
   * a t utilise pour dfinir le dplacement de la particule.
   * </p>
   */
  public void switchKineticComponent()
  {
    SKineticComponent tmp = this.component;
    
    this.component = this.predicted_component;
    this.predicted_component = tmp;
  }
  
  /**
   * Mthode pour dterminer si deux particules possdent le mme numro d'identification (id).
   * 
   * @param p La particule.
   * @return <b>true</b> si les deux particules ont le mme numro d'identification et <b>false</b> sinon.
   */
  public boolean asSameID(SParticle p)
  {
    return this.id == p.id;
  }
    
  @Override
  public String toString() 
  {
    return "SParticle [id=" + id + ", mass=" + mass + ", electric_charge=" + electric_charge + ", component=" + component + "]";
  }

  @Override
  public int hashCode() {
    return Objects.hash(component, electric_charge, mass);
  }

  @Override
  public boolean equals(Object obj)
  {
    if (this == obj)
      return true;
    
    if (obj == null)
      return false;
    
    if (getClass() != obj.getClass())
      return false;
    
    //------------------------------------------------------------------------------------------------------------------------------
    // IMPORTANT : Puisque les particules possdent un numrio d'identification unique, ce paramtre n'est pas test dans le equals.
    //------------------------------------------------------------------------------------------------------------------------------
    SParticle other = (SParticle) obj;
    return Objects.equals(component, other.component)
        && Double.doubleToLongBits(electric_charge) == Double.doubleToLongBits(other.electric_charge)
        && Double.doubleToLongBits(mass) == Double.doubleToLongBits(other.mass);
  }

  
  
}
